// DasmBaseX86.cpp: implementation of the CDasmBaseX86 class.
//
//////////////////////////////////////////////////////////////////////

// Not convinced on operand sizing. look at movzx

#include "stdafx.h"
#include "forge.h"
#include "DasmBaseX86.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif


#define	MAX_INSTRUCTION_LENGTH	16

const char*		szRep[4] =			{ "lock " , "" , "repnz ", "rep "	} ;
const char*		szRepZ[4] =			{ "lock " , "" , "repnz ", "repz "	} ;

const char*		szOperandQualif[5] =	{	"" , "byte ptr" , "word ptr" , "dword ptr" ,"qword ptr" } ;

const char*		szConditions[16] =	{	"o" , "no" , "b" , "nb" , "e" , "ne" , "be" , "a" ,
										"s" , "ns" , "p" , "np" , "l" , "nl" , "le" , "g" } ;

const char*		szMmxPack[4] =		{ "b" , "w" , "d" , "q"	} ;


const char*		szSegReg[8] =	{	"es" , "cs" , "ss" , "ds" , "fs" , "gs" , 0 , 0 } ;
const char*		szMMXReg[8] =	{	"mm0" , "mm1"  , "mm2" , "mm3" , "mm4" , "mm5" , "mm6" , "mm7" } ;
const char*		szFPUReg[8] =	{	"st(0)" , "st(1)"  , "st(2)" , "st(3)" , "st(4)" , "st(5)" , "st(6)" , "st(7)" } ;
const char*		szGenReg[4][8] ={	{ 0, 0 , 0 , 0 , 0 , 0 , 0 , 0 },
									{ "al" , "cl" , "dl" , "bl" , "ah" , "ch" , "dh" , "bh" },
									{ "ax" , "cx" , "dx" , "bx" , "sp" , "bp" , "si" , "di"  },
									{ "eax" , "ecx" , "edx" , "ebx" , "esp" , "ebp" , "esi" , "edi" }
};

//		modxxxrm			->	mod rm
//		General Register	->	grg , g08 , g16 , g32
//		MMX Encodes			->	pp
//		MMX Register		->	xmm
//		Segment Registers	->	2sg 3sg
//		ctrl/dbg Registers	->	_cr / _dr
//		FPU Register		->	fpu
//		Accum. EAX Register ->	eax
//
//		flags : s , w, d
//
//		Conditions			->	tttn
//
//		Values : immediates	-> i08 , is8 , i16 , i32 , i00(default type)
//				 call				c00 
//				 jump				j08 , j16(unused) , j32
//				 direct memadd		a32
//		sep octet		-> : 
//		oos / oas		-> : override operand size  
//		nos / nas		-> : override operand size  


OPSRC	OpCodeMap[] = 
{
	{ "0011 0111:"									, "aaa" , 0 } ,

	{ "1101 0101 : 0000 1010:"						, "aad" , 0 } ,

	{ "1101 0100 : 0000 1010:"						, "aam" , 0 } ,

	{ "0011 1111:"									, "aas" , 0 } ,

	{ "0001 00dw : 11 grg grg:"						, "adc %1 , %0" , 0 } , 
	{ "0001 00dw : mod grg rm"						, "adc %p %1 , %0" , 0 } , 
	{ "1000 00sw : 11 010 grg : i00"				, "adc %0 , %1" , 0 } , 
	{ "0001 010w eax : i00"							, "adc %0 , %1" , 0 } , 
	{ "1000 00sw : mod 010 rm i00"					, "adc %p %0 , %1" , 0 } ,

	{ "0000 00dw : 11 grg grg:"						, "add %1 , %0" , 0 } , 
	{ "0000 00dw : mod grg rm"						, "add %p %1 , %0" , 0 } , 
	{ "1000 00sw : 11 000 grg : i00"				, "add %0 , %1" , 0 } , 
	{ "0000 010w eax: i00"							, "add %0 , %1" , 0 } , 
	{ "1000 00sw : mod 000 rm i00"					, "add %p  %0 , %1" , 0 } , 

	{ "0010 00dw : 11 grg grg:"						, "and %1 , %0" , 0 } , 
	{ "0010 00dw : mod grg rm"						, "and %p %1 , %0" , 0 } , 
	{ "1000 00sw : 11 100 grg: i00"					, "and %0 , %1" , 0 } , 
	{ "0010 010w eax: i00"							, "and %0 , %1" , 0 } , 
	{ "1000 00sw : mod 100 rm i00"					, "and %p %0 , %1" , 0 } , 
	
	{ "0110 0010 : mod grg rm"						, "bound %0 , %1" , 0 } , 

	{ "0000 1111 : 1011 1100 : 11 grg grg:"			, "bsf %1 , %0" , 0 } , 
	{ "0000 1111 : 1011 1100 : mod grg rm"			, "bsf %p %1 , %0" , 0 } , 

	{ "0000 1111 : 1011 1101 : 11 grg grg:"			, "bsr %1 , %0" , 0 } , 
	{ "0000 1111 : 1011 1101 : mod grg rm"			, "bsr %p %1 , %0" , 0 } , 

	{ "0000 1111 : 1100 1 grg:"						, "bswap %0" , 0 } , 

	{ "0000 1111 : 1011 1010 : 11 100 grg : i08"	, "bt %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1010 : mod 100 rm i08"		, "bt %p %0 , %1" , 0 } , 
	{ "0000 1111 : 1010 0011 : 11 grg grg:"			, "bt %1 , %0" , 0 } , 
	{ "0000 1111 : 1010 0011 : mod grg rm"			, "bt %p %1 , %0" , 0 } , 

	{ "0000 1111 : 1011 1010 : 11 111 grg : i08"	, "btc %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1010 : mod 111 rm   i08"	, "btc %p %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1011 : 11 grg grg:"			, "btc %1 , %0" , 0 } , 
	{ "0000 1111 : 1011 1011 : mod grg rm"			, "btc %p %1 , %0" , 0 } , 

	{ "0000 1111 : 1011 1010 : 11 110 grg : i08"	, "btr %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1010 : mod 110 rm i08"		, "btr %p %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 0011 : 11 grg grg:"			, "btr %1 , %0" , 0 } ,
	{ "0000 1111 : 1011 0011 : mod grg rm"			, "btr %p %1 , %0" , 0 } , 
	
	{ "0000 1111 : 1011 1010 : 11 101 grg : i08"	, "bts %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1010 : mod 101 rm i08"		, "bts %p %0 , %1" , 0 } , 
	{ "0000 1111 : 1010 1011 : 11 grg grg:"			, "bts %1 , %0" , 0 } , 
	{ "0000 1111 : 1010 1011 : mod grg rm"			, "bts %p %1 , %0" , 0 } ,

	{ "1110 1000:c32"								, "call %0",	DIT_CALL | DIT_BTVALID } ,		// call direct
	{ "1111 1111:11 010 grg:"						, "call %0",	DIT_CALL | DIT_BTOP0   } ,		// call indirect
	{ "1111 1111:mod 010 rm"						, "call %p %0", DIT_CALL | DIT_BTOPQ0  } , 

	{ "1001 1000:" , "cbw" , 0 } ,
	{ "1001 1001:" , "cdq" , 0 } ,
	{ "1111 1000:" , "clc" , 0 } ,
	{ "1111 1100:" , "cld" , 0 } ,
	{ "1111 1010:" , "cli" , 0 } ,
	{ "1111 0101:" , "cmc" , 0 } ,

	{ "0000 1111: 0100 tttn : 11 grg grg:"		, "cmov%c %0 , %1" , 0 } ,
	{ "0000 1111: 0100 tttn : mod grg rm"		, "cmov%c %0 , %p %1" , 0 } ,

	{ "0011 10dw : 11 grg grg:"					, "cmp %1 , %0" , 0 } ,
	{ "0011 10dw : mod grg rm"					, "cmp %p %1 , %0" , 0 } ,
	{ "1000 00sw : 11 111 grg : i00 "			, "cmp %0 , %1" , 0 } ,
	{ "0011 110w eax : i00"						, "cmp %0 , %1" , 0 } ,
	{ "1000 00sw : mod 111 rm i00"				, "cmp %p %0 , %1" , 0 } ,

	{ "1010 0110:"								, "%rzcmpsb" , 0 } ,
//	{ "oas 1010 0111:"							, "%rzcmpsw" , 0 } ,
//	{ "nas 1010 0111:"							, "%rzcmpsd" , 0 } ,

	{ "0000 1111 : 1011 000w : 11 grg grg:"		, "cmpxchg %1 , %0" , 0 } ,
	{ "0000 1111 : 1011 000w : mod grg rm"		, "cmpxchg %p %1 , %0" , 0 } ,

	{ "0000 1111 : 1100 0111 : a32"				, "cmpxchg8b qword ptr %0" , 0 } ,

	{ "0000 1111 : 1010 0010:"					, "cpuid" , 0 } , 
	{ "1001 1001:"								, "cwd" , 0 } , 
	{ "1001 1000:"								, "cwde" , 0 } , 
	{ "0010 0111:"								, "daa" , 0 } , 
	{ "0010 1111:"								, "das" , 0 } , 

	{ "1111 111w : 11 001 grg:"					, "dec %0" , 0 } , 
	{ "0100 1 grg:"								, "dec %0" , 0 } , 
	{ "1111 111w : mod 001 rm"					, "dec %p %0" , 0 } , 

	{ "1111 011w : 11 110 grg:"					, "div %0" , 0 } , 
	{ "1111 011w : mod 110 rm"					, "div %p %0" , 0 } ,

	{ "1100 1000 : i16 i08"						, "enter %0 , %1" , 0 } , 
	
	{ "1111 011w : 11 111 grg:"					, "idiv %0" , 0 } , 
	{ "1111 011w : mod 111 rm"					, "idiv %p %0" , 0 } , 
	
	{ "1111 011w : 11 101 grg:"					, "imul %0" , 0 } ,
	{ "1111 011w : mod 101 grg:"				, "imul %p %0" , 0 } , 
	{ "0000 1111 : 1010 1111 : 11 grg grg:"		, "imul %0 , %1" , 0 } , 
	{ "0000 1111 : 1010 1111 : mod grg rm"		, "imul %0 , %p %1" , 0 } , 
	{ "0110 10s1 : 11 grg grg: i00 "			, "imul %0 , %1 , %2" , 0 } , 
	{ "0110 10s1 : mod grg rm  i00 "			, "imul %0 , %p %1 , %2" , 0 } ,

	{ "1111 111w : 11 000 grg:"	, "inc %0" , 0 } , 
	{ "0100 0 grg:"				, "inc %0" , 0 } , 
	{ "1111 111w : mod 000 rm"	, "inc %p %0" , 0 } , 
	
	{ "1100 1101 : i08" , "int %0" , 0 } , 
	{ "1100 1100:"		, "int 3" , 0 } , 
	
	{ "1100 1110:"							, "into" , 0 } , 
	{ "1100 1111:"							, "iret"			, DIT_BADFLOW } , 

	{ "0111 tttn : j08"						, "j%c %0"			, DIT_CJMP | DIT_BTVALID } , 
	{ "0000 1111 : 1000 tttn : j32 "		, "j%c %0"			, DIT_CJMP | DIT_BTVALID } , 

//	{ "oas 1110 0011 : j08"					, "jcxz %0" , 0 } , 
//	{ "nas 1110 0011 : j08"					, "jecxz %0" , 0 } , 

	{ "1110 1011 : j08"						, "jmp %0"			, DIT_JUMP | DIT_BTVALID } , 
	{ "1110 1001 : j32"						, "jmp %0"			, DIT_JUMP | DIT_BTVALID } , 
	{ "1111 1111 : 11 100 grg:"				, "jmp %0"			, DIT_JUMP | DIT_BTOP0 } , 
	{ "1111 1111 : mod 100 rm"				, "jmp %p %0"		, DIT_JUMP | DIT_BTOPQ0 } , 

	{ "1110 1010 : i32 i16 "				, "jmp far %1 : %0" , DIT_JUMP | DIT_BTUNKNOWN } ,
	{ "1111 1111 : mod 101 rm"				, "jmp far %p %0 "	, DIT_JUMP | DIT_BTUNKNOWN } , 
	
	{ "1001 1111:" , "lahf" , 0 } , 
	
	{ "1100 0101 : mod grg rm"				, "lds %0" , 0 } , 
	{ "1000 1101 : mod grg rm"				, "lea %0 , %p %1" , 0 } , 

	{ "1100 1001:"							, "leave" , 0 } , 
	{ "1100 0100 : mod grg rm"				, "les %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 0100 : mod grg rm"	, "lfs %0 , %1" , 0 } , 
 
	{ "0000 1111 : 1011 0101 : mod grg rm"	, "lgs %0 , %p %1" , 0 } , 

	{ "1010 1100:"		, "%r_lodsb" , 0 } , 
//	{ "oas1010 1101:"	, "%r_lodsw" , 0 } , 
//	{ "nas1010 1101:"	, "%r_lodsd" , 0 } , 

	{ "1110 0010 : j08" , "loop %0" , 0 } , 
	{ "1110 0001 : j08" , "loope %0" , 0 } , 

	{ "0000 1111 : 1011 0010 : mod grg rm" , "lss %0 , %p %1" , 0 } , 

	{ "1000 10dw:11 grg grg: "		, "mov %1 , %0" , 0 } ,	// mov reg reg
	{ "1000 10dw:mod grg rm"		, "mov %p %1 , %0" , 0 } ,	// mov reg modm
	{ "1100 011w:11 000 grg:i00"	, "mov %0 , %1" , 0 } ,	// mov reg , imm
	{ "1100 011w:mod 000 rm i00"	, "mov %p %0 , %1" , 0 } ,	// mov rm , imm
	{ "1011 w grg:i00"				, "mov %0 , %1" , 0 } ,	// mov reg , imm alternate
	
	{ "1010 00dw eax:a32"			, "mov %0 , %p %1" , 0 } ,	// mov eax , disp

	{ "1000 11d0 : 11 3sg grg :"			, "mov %1 , %0" , 0 } , 
	{ "1000 11d0 : mod 3sg rm"				, "mov %1 , %0" , 0 } , 
	
	{ "1010 0100:"		, "%r_movsb" , 0 } , 
//	{ "oas 1010 0101:"	, "%r_movsw" , 0 } , 
//	{ "nas 1010 0101:"	, "%r_movsd" , 0 } , 

	{ "0000 1111 : 1011 1110 : 11 g32 g08:"	, "movsx %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1111 : 11 g32 g16:"	, "movsx %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 1110 : mod g32 rm"	, "movsx %0 ,byte ptr %1" , 0 } , 
	{ "0000 1111 : 1011 1111 : mod g32 rm"	, "movsx %0 ,word ptr %1" , 0 } , 

	{ "0000 1111 : 1011 0110 : 11 g32 g08:"	, "movzx %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 0111 : 11 g32 g16:"	, "movzx %0 , %1" , 0 } , 
	{ "0000 1111 : 1011 0110 : mod g32 rm"	, "movzx %0 ,byte ptr %1" , 0 } , 
	{ "0000 1111 : 1011 0111 : mod g32 rm"	, "movzx %0 ,word ptr %1" , 0 } , 

	{ "1111 011w : 11 100 grg:"		, "mul %0" , 0 } ,
	{ "1111 011w : mod 100 rm"		, "mul %p %0" , 0 } , 

	{ "1111 011w : mod 011 grg:"	, "neg %0" , 0 } , 
	{ "1111 011w : mod 011 rm"		, "neg %p %0" , 0 } , 

	{ "1001 0000:"					, "nop" , 0 } , 

	{ "1111 011w : 11 010 grg:" , "not %0" , 0 } , 
	{ "1111 011w : mod 010 rm"	, "not %p %0" , 0 } , 

	{ "0000 10dw : 11 grg grg:"			, "or %1 , %0" , 0 } , 
	{ "0000 10dw : mod grg rm"			, "or %p %1 , %0" , 0 } , 
	{ "1000 00sw : 11 001 grg : i00"	, "or %0 , %1" , 0 } ,
	{ "0000 110w eax : i00 "			, "or %0 , %1" , 0 } , 
	{ "1000 00sw : mod 001 rm i00"		, "or %p %0 , %1" , 0 } , 

	{ "1000 1111 : 11 000 grg:"		, "pop %0" , 0 } , 
	{ "0101 1 grg:"					, "pop %0" , 0 } , 
	{ "1000 1111 : mod 000 rm"		, "pop %p %0" , 0 } , 

	{ "0001 1111:"					, "pop ds" , 0 } , 
	{ "0000 0111:"					, "pop es" , 0 } , 
	{ "0001 0111:"					, "pop ss" , 0 } , 
	{ "0000 1111 : 1010 0001 :"		, "pop fs" , 0 } , 
	{ "0000 1111 : 1010 1001 :"		, "pop gs" , 0 } ,
	
//	{ "oos 0110 0001:"				, "popa" , 0 } , 
//	{ "nos 0110 0001:"				, "popad" , 0 } , 
//	{ "oos 1001 1101:"				, "popf" , 0 } , 
//	{ "nos 1001 1101:"				, "popfd" , 0 } , 

	{ "1111 1111:11 110 grg:"		, "push %0" },	// push reg 
	{ "0101 0 grg:"					, "push %0" },	// push reg 	
	{ "1111 1111 : mod 110 rm"		, "push %p %0" },	// push mem
	
//	{ "oos 0110 10s0 : i00"			, "pushw %0" },
//	{ "nos 0110 10s0 : i00"			, "pushd %0" },

	{ "0001 1110:"					, "push cs" , 0 } , 
	{ "0001 0110:"					, "push ss" , 0 } , 
	{ "0001 1110:"					, "push ds" , 0 } , 
	{ "0000 0110:"					, "push es" , 0 } , 
	{ "0000 1111 : 1010 0000 :"		, "push fs" , 0 } , 
	{ "0000 1111 : 1010 1000 :"		, "push gs" , 0 } , 

//	{ "oos 0110 0000:"				, "pusha" , 0 } , 
//	{ "nos 0110 0000:"				, "pushad" , 0 } , 
//	{ "oos 1001 1100:"				, "pushf" , 0 } ,
//	{ "nos 1001 1100:"				, "pushfd" , 0 } , 

	{ "1101 000w : 11 010 grg:"		, "rcl %0 , 1" , 0 } , 
	{ "1101 000w : mod 010 rm"		, "rcl %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 010 grg:"		, "rcl %0 , cl" , 0 } , 
	{ "1101 001w : mod 010 rm"		, "rcl %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 010 grg :i08" , "rcl %0 , %1" , 0 } , 
	{ "1100 000w : mod 010 rm  i08"	, "rcl %p %0 , %1" , 0 } , 

	{ "1101 000w : 11 011 grg:"		, "rcr %0 , 1" , 0 } , 
	{ "1101 000w : mod 011 rm"		, "rcr %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 011 grg:"		, "rcr %0 , cl" , 0 } , 
	{ "1101 001w : mod 011 rm"		, "rcr %p %0 , cl" , 0 } ,
	{ "1100 000w : 11 011 grg :i08" , "rcr %0 , %1" , 0 } , 
	{ "1100 000w : mod 011 rm  i08"	, "rcr %p %0 , %1" , 0 } , 

	{ "1100 0011 :"					, "ret" , DIT_RET } , 
	{ "1100 0010 : i16"				, "ret %0" , DIT_RET } , 

	{ "1100 1011 :"					, "ret far" , DIT_RET } ,
	{ "1100 1010 : i16"				, "ret far %0" , DIT_RET } , 

	{ "1101 000w : 11 000 grg:"		, "rol %0 , 1" , 0 } , 
	{ "1101 000w : mod 000 rm"		, "rol %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 000 grg:"		, "rol %0 , cl" , 0 } ,
	{ "1101 001w : mod 000 rm"		, "rol %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 000 grg :i08" , "rol %0 , %1" , 0 } , 
	{ "1100 000w : mod 000 rm  i08"	, "rol %p %0 , %1" , 0 } , 

	{ "1101 000w : 11 001 grg:"		, "ror %0 , 1" , 0 } , 
	{ "1101 000w : mod 001 rm"		, "ror %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 001 grg:"		, "ror %0 , cl" , 0 } , 
	{ "1101 001w : mod 001 rm"		, "ror %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 001 grg :i08" , "ror %0 , %1" , 0 } , 
	{ "1100 000w : mod 001 rm  i08"	, "ror %p %0 , %1" , 0 } ,

	{ "1001 1110 :"					, "sahf" , 0 } , 

	{ "1101 000w : 11 111 grg:"		, "sar %0 , 1" , 0 } , 
	{ "1101 000w : mod 111 rm"		, "sar %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 111 grg:"		, "sar %0 , cl" , 0 } , 
	{ "1101 001w : mod 111 rm"		, "sar %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 111 grg :i08" , "sar %0 , %1" , 0 } , 
	{ "1100 000w : mod 111 rm  i08"	, "sar %p %0 , %1" , 0 } ,
	
	{ "0001 10dw : 11 grg grg:"		, "sbb %1 , %0" , 0 } , 
	{ "0001 10dw : mod grg rm"		, "sbb %1 , %0" , 0 } , 
	{ "1000 00sw : 11 011 grg: i00"	, "sbb %0 , %1" , 0 } ,
	{ "0001 110w eax : i00"			, "sbb %0 , %1" , 0 } , 
	{ "1000 00sw : mod 011 rm i00"	, "sbb %p %0 , %1" , 0 } , 
	
	{ "1010 1110:"		, "%rzscasb" , 0 } , 
//	{ "oas 1010 1111:"	, "%rzscasw" , 0 } , 
//	{ "nas 1010 1111:"	, "%rzscasd" , 0 } , 

	{ "0000 1111 : 1001 tttn : 11 000 g08:"	, "set%c %0" , 0 } , 
	{ "0000 1111 : 1001 tttn : mod 000 rm"	, "set%c byte ptr %0" , 0 } ,  

	{ "1101 000w : 11 100 grg:"		, "shl %0 , 1" , 0 } , 
	{ "1101 000w : mod 100 rm"		, "shl %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 100 grg:"		, "shl %0 , cl" , 0 } ,
	{ "1101 001w : mod 100 rm"		, "shl %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 100 grg :i08" , "shl %0 , %1" , 0 } , 
	{ "1100 000w : mod 100 rm  i08"	, "shl %p %0 , %1" , 0 } ,

	{ "0000 1111 : 1010 0100 : 11 grg grg : i08"	, "shld %1 , %0 , %2" , 0 } , 
	{ "0000 1111 : 1010 0100 : mod grg rm i08"		, "shld %p %1 , %0 , %2" , 0 } , 
	{ "0000 1111 : 1010 0101 : 11 grg grg:"			, "shld %1 , %0 , cl" , 0 } ,
	{ "0000 1111 : 1010 0101 : mod grg rm"			, "shld %p %1 , %0 , cl" , 0 } , 

	{ "1101 000w : 11 101 grg:"		, "shr %0 , 1" , 0 } , 
	{ "1101 000w : mod 101 rm"		, "shr %p %0 , 1" , 0 } , 
	{ "1101 001w : 11 101 grg:"		, "shr %0 , cl" , 0 } ,
	{ "1101 001w : mod 101 rm"		, "shr %p %0 , cl" , 0 } , 
	{ "1100 000w : 11 101 grg :i08" , "shr %0 , %1" , 0 } , 
	{ "1100 000w : mod 101 rm  i08"	, "shr %p %0 , %1" , 0 } ,

	{ "0000 1111 : 1010 1100 : 11 grg grg : i08"	, "shrd %1 , %0 , %2" , 0 } , 
	{ "0000 1111 : 1010 1100 : mod grg rm i08"		, "shrd %p %1 , %0 , %2" , 0 } , 
	{ "0000 1111 : 1010 1101 : 11 grg grg:"			, "shrd %1 , %0 , cl" , 0 } , 
	{ "0000 1111 : 1010 1101 : mod grg rm"			, "shrd %p %1 , %0 , cl" , 0 } , 

	{ "1111 1001 :" , "stc" , 0 } , 
	{ "1111 1101 :" , "std" , 0 } , 
	{ "1111 1011 :" , "sti" , 0 } , 

	{ "1010 1010 :"		, "%r_stosb" , 0 } , 
//	{ "oas1010 1011 :"	, "%r_stosw" , 0 } , 
//	{ "nas1010 1011 :"	, "%r_stosd" , 0 } , 

	{ "0010 10dw : 11 grg grg:"		, "sub %1 , %0" , 0 } , 
	{ "0010 10dw : mod grg rm"		, "sub %p %1 , %0" , 0 } , 
	{ "1000 00sw : 11 101 grg: i00"	, "sub %0 , %1" , 0 } , 
	{ "0010 110w eax : i00"			, "sub %0 , %1" , 0 } , 
	{ "1000 00sw : mod 101 rm i00"	, "sub %p %0 , %1" , 0 } , 

	{ "1000 010w : 11 grg grg:"		, "test %1 , %0" , 0 } , 
	{ "1000 010w : mod grg rm"		, "test %p %1 , %0" , 0 } , 
	{ "1111 011w : 11 000 grg: i00"	, "test %p %0 , %1" , 0 } , 
	{ "1010 100w eax: i00"			, "test %0 , %1" }, 
	{ "1111 011w : mod 000 rm i00"	, "test %p %0 , %1" , 0 } , 

	{ "0000 1111 : 0000 1011 :"		, "ud2" , 0 } , 

	{ "1001 1011 :" , "wait" , 0 } , 

	{ "0000 1111 : 1100 000w : 11 grg grg:" , "xadd %1 , %0" , 0 } , 
	{ "0000 1111 : 1100 000w : mod grg rm" , "xadd %p %1 , %0" , 0 } , 

	{ "1000 011w : 11 grg grg:"		, "xchg %1 , %0" , 0 } ,
	{ "1000 011w : mod grg rm"		, "xchg %p %1 , %0" , 0 } ,
	{ "1001 0 eax grg:"				, "xchg %0 , %1 " , 0 } ,
	{ "1101 0111:"					, "xlat" },
	{ "0011 00dw : 11 grg grg:"		, "xor %1 , %0" , 0 } ,
	{ "0011 00dw : mod grg rm"		, "xor %p %1 , %0" , 0 } ,
	{ "1000 00sw : 11 110 grg: i00 ", "xor %0 , %1" , 0 } ,
	{ "0011 010w eax: i00"			, "xor %0 , %1" , 0 } ,
	{ "1000 00sw : mod 110 rm i00"	, "xor %p %0 , %1" , 0 } ,

	// ======== Floating Point Instructions

	{ "11011 001 : 1111 0000:"				, "f2xm1" , 0 } ,
	{ "11011 001 : 1110 0001:"				, "fabs" , 0 } ,

	{ "11011 000 : mod 000 rm"				, "fadd dword ptr %0" , 0 } ,
	{ "11011 100 : mod 000 rm"				, "fadd qword ptr %0" , 0 } ,
	{ "11011 000 : 11 000 fpu:"				, "fadd st(0) , %0" , 0 } ,
	{ "11011 100 : 11 000 fpu:"				, "fadd %0 , st(0)" , 0 } ,

	{ "11011 110 : 11 000 fpu:"				, "faddp %0 , st(0)" , 0 } ,
	{ "11011 111 : mod 100 rm"				, "fbld" , 0 } ,
	{ "11011 111 : mod 110 rm"				, "fbstp" , 0 } ,
	{ "11011 001 : 1110 0000:"				, "fchs" , 0 } ,
	{ "11011 011 : 1110 0010:"				, "fclex" , 0 } ,

	{ "11011 010 : 11 000 fpu:"				, "fcmovb st(0) , %0" , 0 } ,
	{ "11011 010 : 11 001 fpu:"				, "fcmove st(0) , %0" , 0 } ,
	{ "11011 010 : 11 010 fpu:"				, "fcmovbe st(0) , %0" , 0 } ,
	{ "11011 010 : 11 011 fpu:"				, "fcmovu st(0) , %0" , 0 } ,
	{ "11011 011 : 11 000 fpu:"				, "fcmovnb st(0) , %0" , 0 } ,
	{ "11011 011 : 11 001 fpu:"				, "fcmovne st(0) , %0" , 0 } ,
	{ "11011 011 : 11 010 fpu:"				, "fcmovnbe st(0) , %0" , 0 } ,
	{ "11011 011 : 11 011 fpu:"				, "fcmovnu st(0) , %0" , 0 } ,

	{ "11011 000 : mod 010 rm"				, "fcom dword ptr %0" , 0 } ,
	{ "11011 100 : mod 010 rm"				, "fcom qword ptr %0" , 0 } ,
	{ "11011 000 : 11 010 fpu:"				, "fcom %0" , 0 } ,

	{ "11011 000 : mod 011 rm"				, "fcomp dword ptr %0" , 0 } ,
	{ "11011 100 : mod 011 rm"				, "fcomp qword ptr %0" , 0 } ,
	{ "11011 000 : 11 011 fpu:"				, "fcomp %0" , 0 } ,

	{ "11011 110 : 11 011 001:"				, "fcompp" , 0 } ,
	{ "11011 011 : 11 110 fpu:"				, "fcomi %0" , 0 } ,
	{ "11011 111 : 11 110 fpu:"				, "fcomip" , 0 } ,
	{ "11011 001 : 1111 1111:"				, "fcos" , 0 } ,
	{ "11011 001 : 1111 0110:"				, "fdecstp" , 0 } ,

	{ "11011 000 : mod 110 rm"				, "fdiv dword ptr %0" , 0 } , 
	{ "11011 100 : mod 110 rm"				, "fdiv qword ptr %0" , 0 } , 	
	{ "11011 000 : 11110 fpu:"				, "fdiv st(0) , %0" , 0 } , 
	{ "11011 100 : 11111 fpu:"				, "fdiv %0 , st(0)" , 0 } , 
	{ "11011 110 : 1111 1 fpu:"				, "fdivp %0 , st(0)" , 0 } , 

	{ "11011 000 : mod 111 rm"				, "fdivr dword ptr %0" , 0 } , 
	{ "11011 100 : mod 111 rm"				, "fdivr qword ptr %0" , 0 } , 
	{ "11011 000 : 11111 fpu:"				, "fdivr st(0) , %0" , 0 } , 
	{ "11011 100 : 11110 fpu:"				, "fdivr %0 , st(0)" , 0 } ,
	{ "11011 110 : 1111 0 fpu:"				, "fdivrp %0  , st(0)" , 0 } , 
	
	{ "11011 101 : 1100 0 fpu:"				, "ffree %0" , 0 } , 
	
	{ "11011 110 : mod 000 rm"				, "fiadd word ptr %0" , 0 } , 
	{ "11011 010 : mod 000 rm"				, "fiadd dword ptr %0" , 0 } , 

	{ "11011 110 : mod 010 rm"				, "ficom word ptr %0" , 0 } , 
	{ "11011 010 : mod 010 rm"				, "ficom dword ptr %0" , 0 } , 

	{ "11011 110 : mod 011 rm"				, "ficomp word ptr %0" , 0 } , 
	{ "11011 010 : mod 011 rm"				, "ficomp dword ptr %0" , 0 } , 

	{ "11011 110 : mod 110 rm"				, "fidiv word ptr %0" , 0 } , 
	{ "11011 010 : mod 110 rm"				, "fidiv dword ptr %0" , 0 } , 
	
	{ "11011 110 : mod 111 rm"				, "fidivr word ptr %0" , 0 } , 
	{ "11011 010 : mod 111 rm"				, "fidivr dword ptr %0" , 0 } , 

	{ "11011 111 : mod 000 rm"				, "fild word ptr %0 " , 0 } , 
	{ "11011 011 : mod 000 rm"				, "fild dword ptr %0 " , 0 } , 
	{ "11011 111 : mod 101 rm"				, "fild qword ptr %0 " , 0 } , 
	
	{ "11011 110 : mod 001 rm"				, "fimul word ptr %0" , 0 } , 
	{ "11011 010 : mod 001 rm"				, "fimul dword ptr %0" , 0 } , 

	{ "11011 001 : 1111 0111:"				, "fincstp" , 0 } , 
	{ "10011011 : 11011011 : 11100011:"		, "finit" , 0 } , 
	{ "11011011 : 11100011:"				, "fninit" , 0 } , 
	
	{ "11011 111 : mod 010 rm"				, "fist word ptr %0" , 0 } , 
	{ "11011 011 : mod 010 rm"				, "fist dword ptr %0" , 0 } , 
	
	{ "11011 111 : mod 011 rm"				, "fistp word ptr %0" , 0 } , 
	{ "11011 011 : mod 011 rm"				, "fistp dword ptr %0" , 0 } , 
	{ "11011 111 : mod 111 rm"				, "fistp qword ptr %0" , 0 } , 
	
	{ "11011 110 : mod 100 rm"				, "fisub word ptr %0" , 0 } , 
	{ "11011 010 : mod 100 rm"				, "fisub dword ptr %0" , 0 } , 

	{ "11011 110 : mod 101 rm"				, "fisubr word ptr %0" , 0 } , 
	{ "11011 010 : mod 101 rm"				, "fisubr dword ptr %0" , 0 } , 

	{ "11011 001 : mod 000 rm"				, "fld dword ptr %0" , 0 } , 
	{ "11011 101 : mod 000 rm"				, "fld qword ptr %0" , 0 } , 
	{ "11011 011 : mod 101 rm"				, "fld tbyte ptr %0" , 0 } , // a voir
	{ "11011 001 : 11 000 fpu:"				, "fld %0" },

	{ "11011 001 : 1110 1000:"				, "fld1" , 0 } , 
	{ "11011 001 : mod 101 rm"				, "fldcw %0" , 0 } , 
	{ "11011 001 : mod 100 rm"				, "fldenv %0" , 0 } , 
	{ "11011 001 : 1110 1010:"				, "fldl2e" , 0 } , 
	{ "11011 001 : 1110 1001:"				, "fldl2t" , 0 } , 
	{ "11011 001 : 1110 1100:"				, "fldlg2" , 0 } , 
	{ "11011 001 : 1110 1101:"				, "fldln2" , 0 } , 
	{ "11011 001 : 1110 1011:"				, "fldpi" , 0 } , 
	{ "11011 001 : 1110 1110:"				, "fldz" , 0 } , 

	{ "11011 000 : mod 001 rm"				, "fmul dword ptr %0" , 0 } , 
	{ "11011 100 : mod 001 rm"				, "fmul qword ptr %0" , 0 } , 
	{ "11011 000 : 11 001 fpu:"				, "fmul st(0) , %0" , 0 } , 
	{ "11011 100 : 11 001 fpu:"				, "fmul %0 , st(0)" , 0 } , 
	{ "11011 110 : 11 001 fpu:"				, "fmulp %0 , st(0)" , 0 } , 

	{ "11011 001 : 1101 0000:"				, "fnop" , 0 } , 
	{ "11011 001 : 1111 0011:"				, "fpatan" , 0 } , 
	{ "11011 001 : 1111 1000:"				, "fprem" , 0 } , 
	{ "11011 001 : 1111 0101:"				, "fprem1" , 0 } , 
	{ "11011 001 : 1111 0010:"				, "fptan" , 0 } , 
	{ "11011 001 : 1111 1100:"				, "frndint" , 0 } , 
	{ "11011 101 : mod 100 rm"				, "frstor %0" , 0 } , 
	{ "11011 101 : mod 110 rm"				, "fsave %0" , 0 } , 
	{ "11011 001 : 1111 1101:"				, "fscale" , 0 } , 
	{ "11011 001 : 1111 1110:"				, "fsin" , 0 } , 
	{ "11011 001 : 1111 1011:"				, "fsincos" , 0 } , 
	{ "11011 001 : 1111 1010:"				, "fsqrt" , 0 } , 
	
	{ "11011 001 : mod 010 rm"				, "fst dword ptr %0" , 0 } , 
	{ "11011 101 : mod 010 rm"				, "fst qword ptr %0" , 0 } , 
	{ "11011 101 : 11 010 fpu:"				, "fst %0" , 0 } , 

	{ "11011 001 : mod 111 rm"				, "fstcw %0" , 0 } , 
	{ "11011 001 : mod 110 rm"				, "fstenv %0" , 0 } , 
	
	{ "11011 001 : mod 011 rm"				, "fstp dword ptr %0" , 0 } , 
	{ "11011 101 : mod 011 rm"				, "fstp qword ptr %0" , 0 } , 
	{ "11011 011 : mod 111 rm"				, "fstp tbyte ptr %0" , 0 } , // a voir
	{ "11011 101 : 11 011 fpu:"				, "fstp %0" , 0 } , 

	{ "11011 111 : 1110 0000:"				, "fstsw ax" , 0 } , 
	{ "11011 101 : mod 111 rm"				, "fstsw word ptr %0" , 0 } , 

	{ "11011 000 : mod 100 rm"				, "fsub dword ptr %0" , 0 } , 
	{ "11011 100 : mod 100 rm"				, "fsub qword ptr %0" , 0 } , 	
	{ "11011 000 : 11100 fpu:"				, "fsub st(0) , %0" , 0 } , 
	{ "11011 100 : 11101 fpu:"				, "fsub %0 , st(0)" , 0 } ,
	{ "11011 110 : 11101 fpu:"				, "fsubp %0 , st(0)" , 0 } ,

	{ "11011 000 : mod 101 rm"				, "fsubr dword ptr %0" , 0 } ,
	{ "11011 100 : mod 101 rm"				, "fsubr qword ptr %0" , 0 } ,
	{ "11011 000 : 11101 fpu:"				, "fsubr st(0) , %0" , 0 } ,
	{ "11011 100 : 11100 fpu:"				, "fsubr %0 , st(0)" , 0 } ,
	{ "11011 110 : 11100 fpu:"				, "fsubrp %0  , st(0)" , 0 } , 	
		
	{ "11011 001 : 1110 0100:"				, "ftst" , 0 } , 
	{ "11011 101 : 1110 0 fpu:"				, "fucom %0" , 0 } , 
	{ "11011 101 : 1110 1 fpu:"				, "fucomp %0" , 0 } , 
	{ "11011 010 : 1110 1001:"				, "fucompp" , 0 } , 
	{ "11011 011 : 11 101 fpu:"				, "fucomi" , 0 } , 
	{ "11011 111 : 11 101 fpu:"				, "fucomip" , 0 } , 
	{ "11011 001 : 1110 0101:"				, "fxam" , 0 } , 
	{ "11011 001 : 1100 1fpu:"				, "fxch %0" , 0 } , 
	{ "11011 001 : 1111 0100:"				, "fxtract" , 0 } , 
	{ "11011 001 : 1111 0001:"				, "fyl2x" , 0 } , 
	{ "11011 001 : 1111 1001:"				, "fyl2xp1" , 0 } , 
	{ "1001 1011:"							, "fwait" , 0 } , 

	// ======== Priviledged Instructions

	{ "0110 0011 : 11 g16 g16:"			, "arpl %1 , %0" , 0 } , 
	{ "0110 0011 : mod g16 rm"			, "arpl %1 , %0" , 0 } , 

	{ "0000 1111 : 0000 0110 :" , "clts" , 0 } , 

	{ "1111 0100:"								, "hlt" , 0 } , 

	{ "1110 010w eax : i08" , "in %0 , %1" , 0 } , 
	{ "1110 110w eax : "	, "in %0 , dx" , 0 } , 

	{ "0110 1100:"			, "%r_insb" , 0 } , 
//	{ "oas 0110 1101:"		, "%r_insw" , 0 } , 
//	{ "nas 0110 1101:"		, "%r_insd" , 0 } , 

	{ "0000 1111 : 0000 1000:"				, "invd" , 0 } , 
	{ "0000 1111 : 0000 0001 : mod 111 rm"	, "invlpg %p %0" , 0 } , 

	{ "0000 1111 : 0000 0010 : 11 grg grg:" , "lar %0 , %1" , 0 } , 
	{ "0000 1111 : 0000 0010 : mod grg rm" , "lar %0 ,%p  %1 " , 0 } , 

	{ "0000 1111 : 0000 0001 : mod 010 rm"	, "lgdt %0" , 0 } ,
	{ "0000 1111 : 0000 0001 : mod 011 rm"	, "lidt %0" , 0 } ,

	{ "0000 1111 : 0000 0000 : 11 010 grg:"	, "lldt %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 010 rm"  , "lldt %p %0" , 0 } , 

	{ "0000 1111 : 0000 0001 : 11 110 grg:"	, "lmsw %0" , 0 } ,
	{ "0000 1111 : 0000 0001 : mod 110 rm"	, "lmsw %p %0" , 0 } , 
	
	{ "1111 0000:"							, "lock" , 0 } , 

	{ "0000 1111 : 0000 0011 : 11 grg grg:" , "lsl %0 , %1" , 0 } , 
	{ "0000 1111 : 0000 0011 : mod grg rm" , "lsl %0 , %p %1" , 0 } , 

	{ "0000 1111 : 0000 0000 : 11 011 grg:" , "ltr %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 011 rm" , "ltr %p %0" , 0 } , 

	{ "0000 1111 : 0010 00d0 : 11 _cr grg:" , "mov %1 , %0" , 0 } ,  
	{ "0000 1111 : 0010 00d1 : 11 _dr grg:"	, "mov %1 , %0" , 0 } , 

	{ "1110 011w eax : i08" , "out %1 , %0" , 0 } , 
	{ "1110 111w eax : "	, "out dx , %0" , 0 } , 
	
	{ "0110 1110:"		, "%r_outsb" , 0 } , 
//	{ "oas 0110 1111:"	, "%r_outsw" , 0 } , 
//	{ "nas 0110 1111:"	, "%r_outsd" , 0 } , 

	{ "0000 1111 : 0011 0010 :"		, "rdmsr" , 0 } , 
	{ "0000 1111 : 0011 0011 :"		, "rdpmc" , 0 } , 
	{ "0000 1111 : 0011 0001 :"		, "rdtsc" , 0 } , 

	{ "0000 1111 : 0000 0001 : mod 000 rm"	, "sgdt %p %0" , 0 } ,

	{ "0000 1111 : 0000 0000 : 11 000 grg:" , "sldt %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 000 rm" , "sldt %0" , 0 } , 

	{ "0000 1111 : 1010 1010: "		, "rsm" , 0 } , 

	{ "0000 1111 : 0000 0000 : 11 001 grg:" , "str %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 001 rm" , "str %0" , 0 } , 

	{ "0000 1111 : 0000 0001 : mod 001 rm" , "sidt %0" , 0 } , 

	{ "0000 1111 : 0000 0001 : 11 100 grg:" , "smsw %0" , 0 } , 
	{ "0000 1111 : 0000 0001 : mod 100 rm" , "smsw %0" , 0 } , 
	
	{ "0000 1111 : 0000 1001 :" , "wbinvd" , 0 } , 

	{ "0000 1111 : 0000 0000 : 11 100 grg:"		, "verr %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 100 rm"		, "verr %p %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : 11 101 grg:"		, "verw %0" , 0 } , 
	{ "0000 1111 : 0000 0000 : mod 101 rm"		, "verw %p %0" , 0 } , 

	{ "0000 1111 : 0011 0000 :" , "wrmsr" , 0 } , 

	// ======== MMX Instructions

	{ "0000 1111 : 01110111: "					, "emms" , 0  } ,

	{ "0000 1111 : 011d1110: 11 xmm grg:"		, "movd %0 , %1" , 0 } , 
	{ "0000 1111 : 011d1110: mod xmm rm"		, "movd %0 , %1" , 0 } , 

	{ "0000 1111 : 011d1111: 11 xmm grg:"		, "movq %0 , %1" , 0 } , 
	{ "0000 1111 : 011d1111: mod xmm rm"		, "movq %0 , %1" , 0 } , 
	
	{ "0000 1111 : 01101011: 11 xmm xmm:"		, "packssdw %0 , %1" , 0 } , 
	{ "0000 1111 : 01101011: mod xmm rm"		, "packssdw %0 , %1" , 0 } ,

	{ "0000 1111 : 01100011: 11 xmm xmm:"		, "packsswb %0 , %1" , 0 } , 
	{ "0000 1111 : 01100011: mod xmm rm"		, "packsswb %0 , %1" , 0 } ,

	{ "0000 1111 : 01100111: 11 xmm xmm:"		, "packuswb %0 , %1" , 0 } , 
	{ "0000 1111 : 01100111: mod xmm rm"		, "packuswb %0 , %1" , 0 } ,
	
	{ "0000 1111 : 111111pp: 11 xmm xmm:"		, "padd%g %0 , %1" , 0 } , 
	{ "0000 1111 : 111111pp: mod xmm rm"		, "padd%g %0 , %1" , 0 } ,

	{ "0000 1111 : 111011pp: 11 xmm xmm:"		, "padds%g %0 , %1" , 0 } , 
	{ "0000 1111 : 111011pp: mod xmm rm"		, "padds%g %0 , %1" , 0 } ,

	{ "0000 1111 : 110111pp: 11 xmm xmm:"		, "paddus%g %0 , %1" , 0 } , 
	{ "0000 1111 : 110111pp: mod xmm rm"		, "paddus%g %0 , %1" , 0 } ,

	{ "0000 1111 : 11011011: 11 xmm xmm:"		, "pand %0 , %1" , 0 } , 
	{ "0000 1111 : 11011011: mod xmm rm"		, "pand %0 , %1" , 0 } ,

	{ "0000 1111 : 11011111: 11 xmm xmm:"		, "pandn %0 , %1" , 0 } , 
	{ "0000 1111 : 11011111: mod xmm rm"		, "pandn %0 , %1" , 0 } ,

	{ "0000 1111 : 011101pp: 11 xmm xmm:"		, "pcmpeq%g %0 , %1" , 0 } , 
	{ "0000 1111 : 011101pp: mod xmm rm"		, "pcmpeq%g %0 , %1" , 0 } ,

	{ "0000 1111 : 011001pp: 11 xmm xmm:"		, "pcmpgt%g %0 , %1" , 0 } , 
	{ "0000 1111 : 011001pp: mod xmm rm"		, "pcmpgt%g %0 , %1" , 0 } ,

	{ "0000 1111 : 11110101: 11 xmm xmm:"		, "pmadd %0 , %1" , 0 } , 
	{ "0000 1111 : 11011111: mod xmm rm"		, "pmadd, %1" , 0 } ,

	{ "0000 1111 : 11100101: 11 xmm xmm:"		, "pmulh %0 , %1" , 0 } ,
	{ "0000 1111 : 11100101: mod xmm rm"		, "pmulh %0 , %1" , 0 } ,

	{ "0000 1111 : 11010101: 11 xmm xmm:"		, "pmull %0 , %1" , 0 } ,
	{ "0000 1111 : 11010101: mod xmm rm"		, "pmull %0 , %1" , 0 } ,

	{ "0000 1111 : 11101011: 11 xmm xmm:"		, "por %0 , %1" , 0 } ,
	{ "0000 1111 : 11101011: mod xmm rm"		, "por %0 , %1" , 0 } ,

	{ "0000 1111 : 111100pp: 11 xmm xmm:"		, "psll%g %0 , %1" , 0 } ,
	{ "0000 1111 : 111100pp: mod xmm rm"		, "psll%g %0 , %1" , 0 } ,
	{ "0000 1111 : 011100pp: 11 100 xmm : i08"	, "psll%g %0 , %1" , 0 } ,

	{ "0000 1111 : 111100pp: 11 xmm xmm:"		, "psra%g %0 , %1" , 0 } ,
	{ "0000 1111 : 111100pp: mod xmm rm"		, "psra%g %0 , %1" , 0 } ,
	{ "0000 1111 : 011100pp: 11 100 xmm : i08"	, "psra%g %0 , %1" , 0 } ,

	{ "0000 1111 : 111100pp: 11 xmm xmm:"		, "psrl%g %0 , %1" , 0 } ,
	{ "0000 1111 : 111100pp: mod xmm rm"		, "psrl%g %0 , %1" , 0 } ,
	{ "0000 1111 : 011100pp: 11 100 xmm : i08"	, "psrl%g %0 , %1" , 0 } ,

	{ "0000 1111 : 111110pp: 11 xmm xmm:"		, "psub%g %0 , %1" , 0 } ,
	{ "0000 1111 : 111110pp: mod xmm rm"		, "psub%g %0 , %1" , 0 } ,

	{ "0000 1111 : 111010pp: 11 xmm xmm:"		, "psubs%g %0 , %1" , 0 } ,
	{ "0000 1111 : 111010pp: mod xmm rm"		, "psubs%g %0 , %1" , 0 } ,

	{ "0000 1111 : 110110pp: 11 xmm xmm:"		, "psubus%g %0 , %1" , 0 } ,
	{ "0000 1111 : 110110pp: mod xmm rm"		, "psubus%g %0 , %1" , 0 } ,

	{ "0000 1111 : 011010pp: 11 xmm xmm:"		, "punpckh%g %0 , %1" , 0 } ,
	{ "0000 1111 : 011010pp: mod xmm rm"		, "punpckh%g %0 , %1" , 0 } ,

	{ "0000 1111 : 011000pp: 11 xmm xmm:"		, "punpckl%g %0 , %1" , 0 } ,
	{ "0000 1111 : 011000pp: mod xmm rm"		, "punpckl%g %0 , %1" , 0 } ,

	{ "0000 1111 : 11101111: 11 xmm xmm:"		, "pxor %0 , %1" , 0 } ,
	{ "0000 1111 : 11101111: mod xmm rm"		, "pxor %0 , %1" , 0 } ,

	{ 0 , 0 , 0 }
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDasmBaseX86::CDasmBaseX86()
{
	m_Buffer = NULL;
	m_BufferSize = 0;
	m_Offset = 0;
	m_TargetBase = 0;
	m_Mode32 = TRUE;		// Default to 32bit mode
}

CDasmBaseX86::~CDasmBaseX86()
{
	Release();
}

// You must supply at least MAX_INSTRUCTION_LENGTH bytes of data for disassembly. (See DasmNext() )

int CDasmBaseX86::Initialise(void *DataSource, int len, int base)
{
	if ( m_Buffer ) Release();
	m_Buffer = malloc( len ); if ( m_Buffer == NULL ) return -1;
	m_BufferSize = len;
	::memmove( m_Buffer, DataSource, len );
	m_Offset = 0;
	m_TargetBase = base;
	return 0;
}

void CDasmBaseX86::Release()
{
	if ( m_Buffer ) free( m_Buffer );
	m_Buffer = NULL;
	m_BufferSize = 0;
	m_Offset = 0;
	m_TargetBase = 0;
}

// ================================================================================================
//
// Output Buffer Control
//
// ================================================================================================

void CDasmBaseX86::InitOPBuffers()
{
	sprintf( m_szOutputBuffer, "<no output from CDasmBaseX86>" );
	sprintf( m_szAddressBuffer, "%08X", HostOffsetToTarget( m_Offset ) );
	sprintf( m_szDataBuffer, "" );
	sprintf( m_szCodeBuffer, "" );
	sprintf( m_szCommentBuffer, "" );
}

void CDasmBaseX86::ConstructOPBuffer()
{
	sprintf( m_szOutputBuffer, "%s : ( %s)", m_szAddressBuffer, m_szDataBuffer );

	while (strlen(m_szOutputBuffer)<48) strcat( m_szOutputBuffer, " " );
	strcat( m_szOutputBuffer, " " );
	strcat( m_szOutputBuffer, m_szCodeBuffer );
	
	while (strlen(m_szOutputBuffer)<80) strcat( m_szOutputBuffer, " " );
	strcat( m_szOutputBuffer, " " );
	if ( strlen( m_szCommentBuffer ) ) 
	{
		strcat( m_szOutputBuffer, "; " );
		strcat( m_szOutputBuffer, m_szCommentBuffer );
	}
}

// ================================================================================================
//
// Address Translation
//
// ================================================================================================

LPVOID CDasmBaseX86::HostPointer()
{
	unsigned char *p = (unsigned char *)m_Buffer;
	p+=m_Offset;
	return (LPVOID)p;
}

int CDasmBaseX86::HostOffsetToTarget(int offset)
{
	return ( offset + m_TargetBase );
}

int CDasmBaseX86::HostToTarget(int host_address)
{
	host_address-=(int)m_Buffer;
	return HostOffsetToTarget( host_address );
}

DWORD CDasmBaseX86::RelToAbs(int o)
{
	// o is in in-line data
	return (DWORD) ( HostOffsetToTarget( m_Offset ) + o );
}

// ================================================================================================
//
// Buffer Access
//
// ================================================================================================

BYTE CDasmBaseX86::PeekByte()
{
	BYTE	*p = (BYTE *)m_Buffer;
	p+=m_Offset;
	return (BYTE)*p;
}

BYTE CDasmBaseX86::GrabByte()
{
	BYTE	b = PeekByte();

//	char	t[16]; sprintf(t, "%02X ", b ); strcat( m_szDataBuffer,t );

	m_Offset++;
	return b;
}

int CDasmBaseX86::GetBit()
{
	if ( m_bitsleft == 0 ) { m_CurByte = GrabByte(); m_bitsleft = 8; }
	m_bitsleft--;
	int rc = m_CurByte & 0x080;
	m_CurByte<<=1;
	if (rc) return 1; 
	return 0;
}

int CDasmBaseX86::GetBits(int n)
{
	int	acc = 0;
	while (n--) 
	{
		acc<<=1;
		acc|=GetBit();
	}
	return acc;
}

// ================================================================================================
//
// Process DATA bytes
//
// ================================================================================================

// Process Next byte as a DATA byte

void CDasmBaseX86::DoDataByte(int count)
{
	while ( count-- )
	{
		BYTE	b = GrabByte();
		char	t[16]; sprintf(t, "%02X", b ); strcat( m_szDataBuffer,t );
	}
}

// ================================================================================================
//
// Instruction Decoding
//
// ================================================================================================

// Scan for potential prefix bytes. If I appear to find > 4 of them return FALSE (assume data)

BOOL CDasmBaseX86::ScanPrefix()
{
	int	iNumFound = 0;
	int	iOrigOffset = m_Offset;

	// Sensible defaults

	m_bAddressSizeOverride	= FALSE;
	m_bOperandSizeOverride	= FALSE;
	m_bLockPrefix			= FALSE;
	m_RepPrefix				= 0;
	m_SegPrefix				= -1;

	while (1)
	{
		if ( iNumFound>4 ) { m_Offset = iOrigOffset; return FALSE; }

		switch( PeekByte() )
		{
		case 0x0F0:				// LOCK:
			m_bLockPrefix = TRUE; iNumFound++; GrabByte(); m_RepPrefix = 0x0F0; break;

		case 0x0F2:				// REPNE/NZ
			m_RepPrefix = 0x0F2; iNumFound++; GrabByte(); break;
		case 0x0F3:				// REPEQ/Z
			m_RepPrefix = 0x0F3; iNumFound++; GrabByte(); break;

		case 0x02E:				// CS:
			m_SegPrefix = 1; iNumFound++; GrabByte(); break;
		case 0x03E:				// DS:
			m_SegPrefix = 3; iNumFound++; GrabByte(); break;
		case 0x026:				// ES:
			m_SegPrefix = 0; iNumFound++; GrabByte(); break;
		case 0x064:				// FS:
			m_SegPrefix = 4; iNumFound++; GrabByte(); break;
		case 0x065:				// GS:
			m_SegPrefix = 5; iNumFound++; GrabByte(); break;
		case 0x036:				// SS:
			m_SegPrefix = 2; iNumFound++; GrabByte(); break;

		case 0x066:				// OpSz
			m_bOperandSizeOverride = TRUE; iNumFound++; GrabByte(); break;
		case 0x067:				// AdSz
			m_bAddressSizeOverride = TRUE; iNumFound++; GrabByte(); break;
		default:
			return TRUE;
		}

	}

}

// Process a SIB from the ModRegRM Field

BOOL CDasmBaseX86::SIBProcess(char *buff)
{
	sprintf( m_szCommentBuffer, "SIBProcess not implemented, bad decode");
	sprintf(buff,"**CANNOT HANDLE SIB**");
	return FALSE;
}

// Process the Mod Reg RM fields to operand buffer, return FALSE if something is invalid.

BOOL CDasmBaseX86::ModMProcess(char *buff)
{
	BOOL bRet = TRUE;						// Be optimistic, assume it will work!
	char *szIdx = buff;
	unsigned int uDisp;

	char szSymbol[ 1024 ];

	if ( m_Mod == 0x03 ) return FALSE;		// Handled elsewhere
	
	// Handle segment override
	if ( m_SegPrefix!=-1 )
	{
		sprintf( szIdx, "%s:" ,szSegReg[m_SegPrefix] ) ;
		szIdx = buff + strlen( buff ) ;						// bump passable string pointer
	}

	// If I need a SIB byte then go somewhere else
	if ( m_RM == 0x04 ) return SIBProcess( szIdx );

	// Process those cases I Can (MOD = 0/1 or 2)
	if( m_Mod == 0x00 )
	{
		if( m_RM == 0x05 )	
		{
			uDisp = *(DWORD*)HostPointer() ;
			m_Offset+=4 ;
			if( m_SegPrefix != -1 )
				sprintf( szIdx ,"[%08X]",uDisp);
			else		
			{
				if ( ResolveSymbol( (LPVOID)uDisp, szSymbol, 1024 ) )
					sprintf( szIdx, "[%s]", szSymbol );
				else
					sprintf( szIdx, "[SYMB_%08X]", uDisp );
			}
		}
		else				// [reg]
			sprintf( szIdx , "[%s]" , szGenReg[3][m_RM] ) ;
	} else {
		if( m_Mod == 0x01 )
		{		// disp8[reg]
			uDisp = (signed int)(*((signed char*)HostPointer()));
			m_Offset++;
			sprintf( szIdx , "[ %s %c %X ]" , szGenReg[3][m_RM], ((int)uDisp<0) ? '-' : '+' , abs((int)uDisp)) ;
		} else {	// disp32[reg]
			uDisp = *(DWORD*)HostPointer();
			m_Offset+=4;
			if ( ResolveSymbol( (LPVOID)uDisp, szSymbol, 1024 ) )
				sprintf( szIdx , "[ %s + %s ]" , szGenReg[3][m_RM], szSymbol );
			else
				sprintf( szIdx , "[ %s + SYMB_%08X ]" , szGenReg[3][m_RM], uDisp );
		}
	}

	// Done

	return bRet;
}

// Attempt to parse current instruction using map provided. If failed then return FALSE.

BOOL CDasmBaseX86::ParseInstruction(char *map )
{
	char *source = map;
	char_buff * current_operand = &m_szOperands[0];
	m_sFlag = 0;
	m_dFlag = 0;
	m_tttnFlag = -1;
	int	dwData;

	int store;
	char t;
	int reg_index;

	BYTE	bSize;
	BOOL	bAddress;
	BOOL	sFlag;

	char	szSymbol[ 1024 ];

	char cSwitchValue;

	while(1)
	{
		cSwitchValue = *map;

		switch( *map++ )
		{
		case '\0':			
			return TRUE;
		case ':':			continue;
		case ' ':			continue;
		case '\t':			continue;

		case '0':	if (GetBit()!=0) { return FALSE; } else continue;
		case '1':	if (GetBit()!=1) { return FALSE; } else continue;

		case 's':	m_sFlag = GetBit(); continue;
		case 'd':	m_dFlag = GetBit(); continue;
		case 'w':	if ( GetBit() == 0 ) m_OperandSize = 1; continue;

		case 't':	m_LastInstructionCond = m_tttnFlag = GetBits(4); map+=3; continue;

		case 'm':	
					store = m_Offset;
					m_Mod = GetBits(2); 
					m_RegOpCode = GetBits(3); 
					m_RM = GetBits(3); 
					m_Offset = store; m_bitsleft = 0; GetBits(2);
					map+=2; 
					continue;

		case 'r':	if (!ModMProcess( *current_operand++ ) ) return FALSE;
					m_bitsleft = 0; map++;
					continue;

		case '2':	strcpy( *current_operand++, szSegReg[ GetBits(2) ] ); map+=2; continue;
		case '3':	strcpy( *current_operand++, szSegReg[ GetBits(3) ] ); map+=3; continue;

		case 'x':	strcpy( *current_operand++, szMMXReg[ GetBits(3) ] ); map+=2; continue;

		case 'f':	strcpy( *current_operand++, szFPUReg[ GetBits(3) ] ); map+=2; continue;

		case 'p':	m_mmxPack = GetBits(2); map++; continue;

		case 'g':	reg_index = GetBits(3); 
					t = *map; map+=2;

					switch (t)
					{
					case 'r':	strcpy( *current_operand++, szGenReg[m_OperandSize][reg_index] ); break;
					case '0':	strcpy( *current_operand++, szGenReg[1][reg_index] ); break;
					case '1':	strcpy( *current_operand++, szGenReg[2][reg_index] ); break;
					case '3':	strcpy( *current_operand++, szGenReg[3][reg_index] ); break;
					default:
						::MessageBox( HWND_DESKTOP, "Illegal register description", "DasmBaseX86 Error", MB_OK );
						return FALSE;
					}

					continue;
		
		case 'e':	strcpy( *current_operand++, szGenReg[m_OperandSize][0] ); map+=2; continue;

		case '_':	::MessageBox( HWND_DESKTOP, "Cannot handle debug or control regs", "XDasm", MB_OK );
					return FALSE;

		case 'o':
		case 'n':	::MessageBox( HWND_DESKTOP, "Cannot handle inbound operand or address size override", "XDasm", MB_OK );
					return FALSE;

		case 'j':
		case 'c':
					if ( *map == '0' )		// 8 bit offset
					{
						dwData = (signed int)*(signed char*)HostPointer();
						m_Offset++;
					} else {				// 32 bit offset
						dwData = *(DWORD*)HostPointer();
						m_Offset+=4;
					}
					map+=2;
		
					// dwData is the Target Rel:address

					m_BranchTarget = (unsigned int)RelToAbs( dwData );

					sprintf( m_szCommentBuffer, "Target Address = %08X", m_BranchTarget );

					if ( ResolveSymbol( (LPVOID)RelToAbs(dwData), szSymbol, 1024 ) )
  						sprintf( *current_operand++, "%s", szSymbol );
					else
  						sprintf( *current_operand++, "SYMBOL_%08X", m_BranchTarget );
					continue;

		case 'a':
		case 'i':
					bAddress = (cSwitchValue=='a');

					if( *(map+1) == '0' )	// if in the for I00 (check last char) then ..
					{
						sFlag = m_sFlag ;			// .. use sbit and operand size from opcode
						bSize = m_OperandSize ;
					} else {						// check for (un)signed 8,16 or 32
						sFlag = ( *map == 's' ) ;
						if( *(map+1) == '8'  ) bSize = 1 ;
						else if( *(map+1) == '6' ) bSize = 2 ;
						else bSize = 3 ;
					};

					map+=2;

					// At this point I know the type (address/immediate) the sign-extension requirement
					// and the data size, so read them!

					if(sFlag)
					{
						dwData = (signed int)*(signed char*)HostPointer();
						m_Offset++;
						sprintf( *current_operand++, "%8X", dwData );
					} else {
						switch(bSize)
						{
						case 1:
								dwData = (int)*(BYTE*)HostPointer();
								m_Offset++;
								sprintf( *current_operand++, "%02X", (unsigned) dwData );
								continue;
						case 2: 
								dwData = (int)*(WORD*)HostPointer();
								m_Offset+=2;
								sprintf( *current_operand++, "%04X", (unsigned) dwData );
								continue;
						case 3:
								dwData = (int)*(DWORD*)HostPointer();
								m_Offset+=4;

								if( bAddress )
								{
	  								if ( ResolveSymbol( (LPVOID)dwData, szSymbol, 1024 ) )
									{
										if( m_SegPrefix != -1 )
											sprintf( *current_operand++, "%s:[%s]" ,szSegReg[m_SegPrefix], szSymbol) ;
										else
											sprintf( *current_operand++, "[%s]" , szSymbol) ;
									} else {
										if( m_SegPrefix != -1 )
											sprintf( *current_operand++, "%s:[%08X]" ,szSegReg[m_SegPrefix], szSymbol) ;
										else
											sprintf( *current_operand++, "[%08X]" , dwData) ;
									}

								} else {
									// should only print "offset" if found as a symbol
	  								if ( ResolveSymbol( (LPVOID)dwData, szSymbol, 1024 ) )
										sprintf( *current_operand++, "offset %s" , szSymbol) ;
									else
										sprintf( *current_operand++, "# %08X" , dwData) ;
								}

								continue;
						default: ::MessageBox( HWND_DESKTOP, "Internal Error#1", "XDASM", MB_OK );
								 return FALSE;
						}
					}
		
					continue;
		default:
			char	ts[256]; sprintf(ts,"Unhandled option '%c' in decode mask for instruction %s", cSwitchValue, map );
			::MessageBox( HWND_DESKTOP, source, ts, MB_OK );
			return FALSE;
		}
	}
	return FALSE;
}

// Walk the opcode table attempting to parse current instruction with each map. 
// On success do the process for it. If you reach the end of the table then the 
// instruction is invalid so return FALSE. (will be treated as data)

BOOL CDasmBaseX86::ScanOpcode()
{

	OPSRC	*table = &OpCodeMap[0];
	while ( table->szDecodeMap )
	{
		int	store = m_Offset;
		m_bitsleft = 0;

		strcpy( m_szOperands[0], "unused" );
		strcpy( m_szOperands[1], "unused" );
		strcpy( m_szOperands[2], "unused" );

		m_LastInstructionInfo = table->info;

		if ( ParseInstruction( table->szDecodeMap ) )
		{
			PrintInstruction( table->szOutputMap, &m_szCodeBuffer[0] );
			return TRUE;
		}
		m_Offset = store;
		table++;
	}
	return	FALSE;
}

// ================================================================================================
//
// Generate human readable output string for current instruction
//
// ================================================================================================

void CDasmBaseX86::PrintInstruction( char *Format, char *Instr )
{
	static char_buff pszTemp ;
	char* pszOut = Instr ;
	const char* pszCurFmt ;
	const char* pszCur ;
	bool bPtrQualif = false;

	for( pszCurFmt = Format ; *pszCurFmt != '\0'; pszCurFmt ++ )
	{
		if( *pszCurFmt == '%' )
		{
			pszCurFmt++ ;
			switch( *pszCurFmt )
			{
				case '0':
				case '1':
				case '2':
					if( bPtrQualif )
					{
						strcpy( pszTemp ,  m_szOperands[ *pszCurFmt - '0']  );
						sprintf( m_szOperands[ *pszCurFmt - '0'] ,"%s %s" , szOperandQualif[ m_OperandSize ] , (const char*)pszTemp) ;
					}
					bPtrQualif = false ;
					break ;
				case 'p':
					bPtrQualif = true ;
					break ;
			}
		}
	}

	// swap op0, op1 based on direction flag
	if( m_dFlag )
	{
		strcpy( pszTemp , m_szOperands[ 0 ] ) ;
		strcpy( m_szOperands[ 0 ] , m_szOperands[ 1 ] ) ;
		strcpy( m_szOperands[ 1 ] , pszTemp ) ;
	}

	for(  pszCurFmt = Format ; *pszCurFmt != '\0'; pszCurFmt ++ )
	{
		if( *pszCurFmt == '%' )
		{
			pszCurFmt++ ;
			switch( *pszCurFmt )
			{
				case '0':
				case '1':
				case '2':
					pszCur = m_szOperands[ *pszCurFmt - '0'] ;
					for(  ; *pszCur != '\0' ; ++pszCur )
						*pszOut++ =	*pszCur	;
					break ;
				case 'p':
					break ;
				case 'c':
					for( pszCur = szConditions[ m_tttnFlag ] ; *pszCur != '\0' ; ++pszCur )
						*pszOut++ =	*pszCur	;
					break ;
				case 'g':
					for( pszCur = szMmxPack[ m_mmxPack ] ; *pszCur != '\0' ; ++pszCur )
						*pszOut++ =	*pszCur	;
					break ;
				case 'r':
					pszCurFmt++ ;
					if( m_RepPrefix != 0 )
					{
						const char* pszPrefix = 0 ;
						if( *pszCurFmt == '_' )
							pszPrefix = szRep[ m_RepPrefix - 0xf0 ] ;
						else if(*pszCurFmt == 'z' )
							pszPrefix = szRepZ[ m_RepPrefix - 0xf0 ] ;
						else
							::MessageBox( HWND_DESKTOP, "Internal Error#2", "XDasm", MB_OK );
						for( pszCur = pszPrefix ; *pszCur != '\0' ; ++pszCur )
							*pszOut++ =	*pszCur	;

					}
					break ;
				default:
					::MessageBox( HWND_DESKTOP, "Internal Error#3", "XDasm", MB_OK );
			}
		}
		else
			*pszOut++ = *pszCurFmt ;
	}
	*pszOut = '\0' ;

	// Tab align operation
//	strcpy( pszTemp, Instr );
//	int done_i = 0;
//	int p = 0;
//	char *s = pszTemp;
//	char *d = Instr;
//	while (*s!='\0')
//	{
//		if ( *s==' ' && done_i == 0 )
//		{
//			while ( p < 10 ) { *d++ = ' '; p++; }
//			s++;
//			done_i = 1;
//		} else {
//			*d++ = *s++;
//			p++;
//		}
//	}
//	*d='\0';
}

// ================================================================================================
//
// Interface
//
// ================================================================================================

// Disassemble next instruction, create output string and return 0. 
// If there is insufficient buffer data to safely perform the disassembly return -1;

int CDasmBaseX86::DasmNext()
{
	int	orig_offset = m_Offset;

	InitOPBuffers();

	if ( m_Offset >= ( m_BufferSize - MAX_INSTRUCTION_LENGTH ) ) return -1;

	if ( PreInstructionProcess() )
	{
		ConstructOPBuffer();
		return 0;
	}

	// If I appear to find too many valid prefix bytes (>4) then I'm in a data section so barf.
	if( !ScanPrefix() ) { DoDataByte(1); return 0; }

	// Set up operand and address sizes based on segment type (m_mode32) and any encounterend prefixes.
	
	m_OperandSize = ( ( m_Mode32 ^ m_bOperandSizeOverride ) == 0 ) ? 2 : 3 ;
	m_AddressSize = ( ( m_Mode32 ^ m_bAddressSizeOverride ) == 0 ) ? 2 : 3 ;

	// Attempt to decode and instruction. If failed then do next byte as data
	int store = m_Offset;
	if ( !ScanOpcode() ) 
	{
		m_Offset = store;
		if ( ProcessDataSection() == 0 ) DoDataByte(1);
	} else {
		int count = m_Offset - orig_offset;
		m_Offset = orig_offset;
		m_LastInstructionLength = count;
		DoDataByte( count );
	}
	ConstructOPBuffer();
	return 0;
}

// A version of DasmNext which puts the output into a CString

int CDasmBaseX86::DasmNextToCString(CString &op)
{
	int	rc = DasmNext();
	op = m_szOutputBuffer;
	return rc;
}

// ================================================================================================
//
// Over-ridables
//
// ================================================================================================

BOOL CDasmBaseX86::ResolveSymbol(LPVOID vTargetAddress, char *szBuffer, int nMaxLen)
{
	return FALSE;
}

BOOL CDasmBaseX86::ProcessDataSection()
{
	return FALSE;
}

BOOL CDasmBaseX86::PreInstructionProcess()
{
	return FALSE;
}

// ================================================================================================
//
// Unclassified Instructions
//
// ================================================================================================


