// DasmBaseX86.h: interface for the CDasmBaseX86 class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_DASMBASEX86_H__98A7E1A0_6440_11D3_90E1_00C0F0219F72__INCLUDED_)
#define AFX_DASMBASEX86_H__98A7E1A0_6440_11D3_90E1_00C0F0219F72__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

typedef struct tagOPSRC
{
	char	*szDecodeMap;
	char	*szOutputMap;
	unsigned int info;
} OPSRC;

// DIT_CALL				Instruction is a call
// DIT_JUMP				Instruction is an unconditional jump
// DIT_CJMP				Instruction is a conditional jump ( m_LastInstructionCond is valid )
// DIT_RET				Instruction is a return
// DIT_BADFLOW			Instruction is a flow-control I haven't handled

// DIT_BTVALID			Branch target for DIT_xxx is decoded already in m_BranchTarget
// DIT_BTOP0			Branch target requires the decode of m_szOperands[0]
// DIT_BTOPQ0			As above except the size qualifier is also required
// DIT_BTUNKNOWN		Cannot handle the required form of branch target yet

#define	DIT_CALL		0x080000000
#define DIT_JUMP		0x040000000
#define DIT_CJMP		0x020000000
#define DIT_RET			0x010000000
#define DIT_BADFLOW		0x008000000
#define DIT_BTVALID		0x000008000
#define DIT_BTOP0		0x000004000
#define DIT_BTOPQ0		0x000002000
#define DIT_BTUNKNOWN	0x000001000

typedef char char_buff[128];

class CDasmBaseX86  
{
public:

	unsigned int m_LastInstructionInfo;			// Info about instruction decoded
	unsigned int m_LastInstructionCond;			// Conditional Index for Last instruction which had one;
	unsigned int m_LastInstructionLength;		// Length in bytes of last instruction.
	unsigned int m_BranchTarget;				// Branch Target for last decoded instruction which had one

	virtual BOOL PreInstructionProcess( void );
	virtual BOOL ProcessDataSection( void );
	DWORD RelToAbs( int o );
	virtual BOOL ResolveSymbol( LPVOID vTargetAddress, char *szBuffer, int nMaxLen );
	CDasmBaseX86();
	virtual ~CDasmBaseX86();

	void PrintInstruction( char *Format, char *Instr );

	BOOL SIBProcess( char *buff );
	BOOL ModMProcess( char *buff );
	LPVOID HostPointer( void );

	int m_mmxPack;
	int m_Mod;
	int m_RegOpCode;
	int m_RM;
	int m_tttnFlag;
	int m_dFlag;				// direction flag			0 = off, 1 = on;
	int m_sFlag;				// sign-extend flag			0 = off, 1 = on;

	char_buff m_szOperands[3];

	BYTE m_CurByte;
	int GetBits( int n );
	int GetBit( void );
	int m_bitsleft;
	BOOL ParseInstruction( char *map );

	BOOL ScanOpcode( void );

	int m_AddressSize;
	int m_OperandSize;

	BYTE GrabByte( void );
	BYTE PeekByte( void );
	void DoDataByte( int count );
	BOOL ScanPrefix( void );
	BOOL m_bLockPrefix;
	BOOL m_bAddressSizeOverride;
	BOOL m_bOperandSizeOverride;
	int m_RepPrefix;						// 0 = invalid, -1 = REPNE, 1 = REPEQ
	int m_SegPrefix;

	BOOL m_Mode32;

	void InitOPBuffers( void );
	void ConstructOPBuffer( void );
	char	m_szOutputBuffer[ 1024 ];
	char	m_szAddressBuffer[ 1024 ];
	char	m_szDataBuffer[ 1024 ];
	char	m_szCodeBuffer[ 1024 ];
	char	m_szCommentBuffer[ 1024 ];

	int HostOffsetToTarget( int offset );
	int HostToTarget( int host_address );
	int DasmNext( void );
	void Release( void );
	virtual int Initialise( void *DataSource, int len, int base );
	int DasmNextToCString( CString& op );
	int m_BufferSize;
	int m_TargetBase;
	int m_Offset;
	void * m_Buffer;

};

#endif // !defined(AFX_DASMBASEX86_H__98A7E1A0_6440_11D3_90E1_00C0F0219F72__INCLUDED_)

