// DasmX86.cpp: implementation of the CDasmX86 class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "forge.h"
#include "DasmX86.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDasmX86::CDasmX86()
{

}

CDasmX86::~CDasmX86()
{

}

//////////////////////////////////////////////////////////////////////

BYTE CDasmX86::cfetch( LPVOID address )
{
	BYTE	rc;
	m_ok = ::ReadProcessMemory( m_hProcess, address, &rc, sizeof(rc), NULL );
	return rc;
}

DWORD CDasmX86::fetch( LPVOID address )
{
	DWORD	rc;
	m_ok = ::ReadProcessMemory( m_hProcess, address, &rc, sizeof(rc), NULL );
	return rc;
}

// Check target address 'a' for a valid NFA (if found then return NFA, otherwise return 0.)
// auto abort is m_ok fails!
//
// Checks for valid name:
//		1)		current byte is non-zero
//		2)		Previous byte is Fx
//		3)		Counted String at current location is zero-terminated
//		4)		Counted String at current contains only printable ascii chars  (33..127 inclusive)
//		5)		((( not done! If an extra check is required the CELL before the previous byte should either
//				    be Zero or a fetchable target location )))

int CDasmX86::NameQuery(LPVOID a)
{
	BYTE	*p = (BYTE*)a;

	int	stringlen = cfetch(a);
	if ( m_ok == FALSE ) return 0;

	if ( stringlen==0 ) return 0;											// TEST#1

	if ( ( cfetch( (LPVOID)(p-1) ) & 0x0F0 ) != 0x0F0 ) return 0;			// Test#2
	if ( m_ok == FALSE ) return 0;

	LPVOID  t = (LPVOID)(p+1+stringlen);								// Test#3
	if ( cfetch( t ) != 0 ) return 0;
	if ( m_ok == FALSE ) return 0;

	LPVOID	n = (LPVOID) (p+1) ; 													// Test#4
	while (stringlen--)
	{
		BYTE	c = cfetch(n);
		if ( m_ok == FALSE ) return 0;
		if ( (c<33) || (c>128) ) return 0;
		BYTE *tt = (BYTE *)n; tt++; n = (LPVOID)tt;
	}

	return (int)a;
}

// A variation on the "classic" IP>NFA Forth definition.
// address			Target address
// cs				CString to write output to
//
// Return NFA on sucess
//
int CDasmX86::IPtoNFA(LPVOID address, int nWindowNeg, int nWindowPos, CString &cs)
{
	int	original = (int)address;
	
	while ( ( NameQuery( address ) == 0 ) && (nWindowNeg--) )
	{
		if (m_ok == FALSE) return FALSE;
		BYTE *p= ( BYTE*) address; p--; address=(LPVOID)p;
	}

	if ( nWindowNeg == 0 ) return 0;

	// address now points at NFA
	// now perform NAME> and check range of XT
	int nfa = (int)address;

//  count + #17 + aligned
//  begin dup c@ $CC = while 1+ repeat \ skip INT3 padding for CREATE

	BYTE	*pp = (BYTE *)address;
	int len = cfetch((LPVOID)pp);
	pp+=(len+18);						// count + #17 +
	while ( (int)pp & 0x03 ) pp++;		// aligned
	while (cfetch( (LPVOID)pp ) == 0x0CC ) pp++;	// begin dup c@ $CC while 1+ repeat 
	address = (LPVOID)pp;
	
	// address is now XT of assumed symbol

	int	disp = (int)original - (int)address;

	// Address in Window
	NameCopy( (LPVOID)nfa, cs );
	if ( disp )
	{
		char tt[256]; 
		if ( disp < 0 )	
			sprintf(tt,"-%x", abs(disp) ); 
		else
			sprintf(tt,"+%x", abs(disp) ); 
		cs+=tt;
	}

	return nfa;
}

BOOL CDasmX86::ResolveSymbol(LPVOID vTargetAddress, char *szBuffer, int nMaxLen)
{
	CString	label;
	if ( IPtoNFA( vTargetAddress, 2048, 0, label ) )
	{
		sprintf( szBuffer, "%s", label );
		return TRUE;
	} else {
		sprintf( szBuffer, "@ %08X", (DWORD)vTargetAddress );
		return FALSE;
	}
}

int CDasmX86::Initialise(void *DataSource, int len, int base, HANDLE hProcess)
{
	m_hProcess = hProcess;
	return CDasmBaseX86::Initialise(DataSource, len, base );
}

void CDasmX86::NameCopy(LPVOID t, CString &cs)
{
	cs = "";
	BYTE *p = (BYTE *)t;

	BYTE len = cfetch( (LPVOID)p++ );
	while (len--)
	{
		BYTE c = cfetch( (LPVOID)p++ );
		cs = cs +(TCHAR)c;
	}
}

// Handle any possible unique handling of a section

// Case1:	Look for possible dictionary header
//			A CELL ALIGNED address
//			1 CELL	(A valid Target address or 0)
//			A Control Byte (0x0Fx)
//			A Valid counted string
//			A 0 Byte
//			4 cells of misc data
//			Some Alignment info (hex 00 20 or CC upto next aligned edge)

BOOL CDasmX86::ForthHeader()
{
	// Check for aligned target address
	int	address = HostOffsetToTarget( m_Offset );
	if ( address & 3 ) return FALSE;

	// Extract Link
	DWORD link = GrabByte() | (GrabByte()<<8 ) | (GrabByte()<<16) | (GrabByte()<<24);
//	if ( link != 0 )
//	{
//		fetch( (LPVOID)link );
//		if (m_ok == FALSE) return FALSE;
//	}

	// Extract control byte
	BYTE control = GrabByte();
	if ( (control & 0x0F8) != 0x0F0 ) return FALSE;

	// Extract counted string information
	BYTE namelen = GrabByte();
	if ( namelen == 0 ) return FALSE;

	char name[257];
	for ( int p=0;p<namelen;p++ ) name[p] = GrabByte(); name[p]=0;
	for ( p=0;p<namelen;p++ ) { if ( (name[p]<33)||(name[p]>127) ) return FALSE; }

	// Check Terminator
	if ( GrabByte() != 0 ) return FALSE;

	// Extract other data fields

	// Extract Line
	DWORD line = GrabByte() | (GrabByte()<<8 ) | (GrabByte()<<16) | (GrabByte()<<24);

	// Extract Xref
	DWORD XRef = GrabByte() | (GrabByte()<<8 ) | (GrabByte()<<16) | (GrabByte()<<24);

	// Extract CodeStream
	DWORD codestream = GrabByte() | (GrabByte()<<8 ) | (GrabByte()<<16) | (GrabByte()<<24);

	// Extract CodeGenerator
	DWORD codesgenerator = GrabByte() | (GrabByte()<<8 ) | (GrabByte()<<16) | (GrabByte()<<24);

	// Align
	while ( (int)HostPointer() & 3 ) GrabByte();

	// Skip any other Int3's I can find ( code-align used for children of create)

	while ( PeekByte()==0x0CC ) GrabByte();

	// Make Output and return

	sprintf( m_szDataBuffer, "Potential Forth Dictionary Header for %s ", name );
	return TRUE;
}

BOOL CDasmX86::PreInstructionProcess()
{
	int	store;
	store = m_Offset; if ( ForthHeader() ) return TRUE; m_Offset = store;

	return FALSE;
}

BOOL CDasmX86::ProcessDataSection()
{
	return FALSE;
}

