// DebugCodeView.cpp: implementation of the CDebugCodeView class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "forge.h"
#include "DebugCodeView.h"

#include "DasmX86.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

IMPLEMENT_DYNAMIC(CDebugCodeView, CSizingControlBar)

BEGIN_MESSAGE_MAP(CDebugCodeView, CSizingControlBar)
	//{{AFX_MSG_MAP(CDebugCodeView)
	ON_WM_SIZE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#define DASM_BUFF_SIZE	256

CDebugCodeView::CDebugCodeView()
{
	m_TargetProcess = NULL;
	m_TargetAddress = NULL;
}

CDebugCodeView::~CDebugCodeView()
{

}

BOOL CDebugCodeView::Create(LPCTSTR lpszWindowName, CWnd * pParentWnd, CSize sizeDefault, BOOL bHasGripper)
{
	CSizingControlBar::Create(lpszWindowName, pParentWnd, sizeDefault, bHasGripper, IDR_DEBUGCODEVIEW );

	m_list.Create( WS_CHILD | WS_VISIBLE | WS_BORDER | WS_HSCROLL | WS_VSCROLL , CRect(0,0,0,0), this, 1 );
	m_Font.CreatePointFont( 100, "Courier", NULL );
	m_list.SetFont(&m_Font);

	return TRUE;
}

void CDebugCodeView::OnSize(UINT nType, int cx, int cy)
{
	if (m_list.m_hWnd)
	{
		m_list.MoveWindow( 0, 0, cx, cy );
	}
}

void CDebugCodeView::OnReset()
{
	LoadState( _T("DEBUGCODEVIEW_PERM") );
	DisplayCode( m_TargetProcess, m_TargetAddress );
}

void CDebugCodeView::OnDebugFinished()
{
	RememberVisibility( IsWindowVisible() );
	SaveState( _T("DEBUGCODEVIEW_PERM") );
}

void CDebugCodeView::DisplayCode( HANDLE hProcess, LPVOID vTarget )
{
	// Display code from vTarget onwards for a bit.

	// If I can find a line for vTarget then scroll to it, otherwise do a refresh
	
	if (FindCodeLine(vTarget)==0) RefreshCode( hProcess, vTarget );
	FindCodeLine(vTarget);
}

void CDebugCodeView::RefreshCode(HANDLE hProcess, LPVOID vTarget)
{
	m_TargetAddress = vTarget;
	if ( (int)hProcess!=-1 ) m_TargetProcess = hProcess;

	if ( m_TargetProcess == NULL ) return;

	LPVOID	buff = GlobalAlloc( GPTR, DASM_BUFF_SIZE );
	::ReadProcessMemory( m_TargetProcess, vTarget, buff, DASM_BUFF_SIZE, NULL );
	m_DasmClass.Initialise( buff, DASM_BUFF_SIZE, (int)vTarget, m_TargetProcess );
	::GlobalFree( buff );

//	BOOL bdid_lock = m_list.LockWindowUpdate();
	m_list.ResetContent();
	CString op;
	while ( m_DasmClass.DasmNextToCString(op) != -1 )
		m_list.AddString( op );
//	if (bdid_lock) m_list.LockWindowUpdate();
}

int CDebugCodeView::FindCodeLine(LPVOID vTarget)
{
	char	t[256];
	sprintf(t,"%08X :", (DWORD)vTarget );
	return ( m_list.SelectString( -1, t ) != LB_ERR );
}
