// DebugConsole.cpp: implementation of the CDebugConsole class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "forge.h"
#include "DebugConsole.h"
#include "MainFrm.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

IMPLEMENT_DYNAMIC(CDebugConsole, CSizingControlBar)

BEGIN_MESSAGE_MAP(CDebugConsole, CSizingControlBar)
	//{{AFX_MSG_MAP(CProjectWindow)
	ON_WM_SIZE()
	ON_EN_MAXTEXT(2, OnEditMaxText)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDebugConsole::CDebugConsole()
{
}

CDebugConsole::~CDebugConsole()
{
}

BOOL CDebugConsole::Create(LPCTSTR lpszWindowName, CWnd * pParentWnd, CSize sizeDefault, BOOL bHasGripper)
{
	CSizingControlBar::Create(lpszWindowName, pParentWnd, sizeDefault, bHasGripper, IDR_DEBUGCONSOLE );
	m_list.Create( WS_CHILD | WS_VISIBLE | WS_BORDER | WS_HSCROLL | WS_VSCROLL | LBS_USETABSTOPS, CRect(0,0,0,0), this, 1 );

	m_Font.CreatePointFont( 100, "Courier", NULL );
	m_list.SetFont(&m_Font);
	m_Edit.Create( WS_CHILD | WS_VISIBLE | WS_BORDER | ES_MULTILINE | ES_WANTRETURN, CRect(0,0,0,0), this, 2 );
	m_Edit.SetFont(&m_Font );

	return TRUE;
}

void CDebugConsole::OnSize(UINT nType, int cx, int cy)
{
	m_Edit.SetFocus();
	if (m_list.m_hWnd)   m_list.MoveWindow( 0, 0, cx, cy-20 );
	if (m_Edit.m_hWnd)   m_Edit.MoveWindow( 0, cy-20, cx, 20 );
}

void CDebugConsole::OnReset()
{
	LoadState( _T("DEBUGCONSOLE_PERM") );
	m_list.ResetContent();
}

void CDebugConsole::OnDebugFinished()
{
	RememberVisibility( IsWindowVisible() );
	SaveState( _T("DEBUGCONSOLE_PERM") );
}

void CDebugConsole::TextOut(CString& cs)
{
	CString t;

	while (cs.GetLength()>0)
	{
		int i = cs.Find( '\n' );
		if ( i == -1 ) i = cs.GetLength();
		t = cs.Left( i ); 
		m_list.AddString( t );

		int j = m_list.GetCount()-2;
		m_list.SelectString( j, t );

		cs = cs.Right( cs.GetLength()-i-1 );
	}

}

void CDebugConsole::OnEditMaxText()
{
	CString	command;

	m_Edit.GetWindowText( command );
	TextOut( command );
	m_Edit.GetWindowText( command );
	m_Edit.SetWindowText( _T("") );

	CString c = command.SpanExcluding( _T(" \t") );
	command = command.Right( command.GetLength() - c.GetLength() );
	command.TrimLeft();

	BOOL	bDone = FALSE;
	if ( c.GetLength() == 0 ) bDone = TRUE;
	if ( c.CompareNoCase( _T("h") ) == 0 )		{ bDone = TRUE; C_Help(); }
	if ( c.CompareNoCase( _T("?") ) == 0 )		{ bDone = TRUE; C_Help(); }
	if ( c.CompareNoCase( _T("help") ) == 0 )	{ bDone = TRUE; C_Help(); }
	if ( c.CompareNoCase( _T("db") ) == 0 )		{ bDone = TRUE; C_DumpMemory( command,1 ); }
	if ( c.CompareNoCase( _T("dw") ) == 0 )		{ bDone = TRUE; C_DumpMemory( command,2 ); }
	if ( c.CompareNoCase( _T("dd") ) == 0 )		{ bDone = TRUE; C_DumpMemory( command,4 ); }
	if ( c.CompareNoCase( _T("x") ) == 0 )		{ bDone = TRUE; C_ContinueExec(); }
	if ( c.CompareNoCase( _T("execute") ) == 0 ){ bDone = TRUE; C_ContinueExec(); }
	if ( c.CompareNoCase( _T("go") ) == 0 )		{ bDone = TRUE; C_ContinueExec(); }
	if ( c.CompareNoCase( _T("user") ) == 0 )	{ bDone = TRUE; C_User(); }
	if ( c.CompareNoCase( _T("dasm") ) == 0 )	{ bDone = TRUE; C_Dasm( command ); }
	if ( c.CompareNoCase( _T("ua") ) == 0 )		{ bDone = TRUE; C_Dasm( command ); }
	if ( c.CompareNoCase( _T("step") ) == 0 )	{ bDone = TRUE; C_Step(); }
	if ( c.CompareNoCase( _T("s") ) == 0 )		{ bDone = TRUE; C_Step(); }
	if (!bDone) { c = _T("Error: Command ")+c; c+=_T(" not recognized."); TextOut(c); }
}

void CDebugConsole::C_DumpMemory(CString &parm, int width)
{
	CMainFrame *mf = (CMainFrame *)AfxGetMainWnd();

	int i = CStringToInt( parm );

	if ( mf->m_wndMemView.IsWindowVisible() == 0 ) mf->ToggleWindowVisibility( &mf->m_wndMemView );
	mf->DebugMemDump( (LPVOID)i, width );
}

void CDebugConsole::C_Help()
{
	CString op = _T("\nCommands");
			op+= _T("\nh ? help          Display this help text.");
			op+= _T("\ndb <address>      Display target memory around \"address\" in BYTE format.");
			op+= _T("\ndw <address>      Display target memory around \"address\" in WORD format.");
			op+= _T("\ndd <address>      Display target memory around \"address\" in DWORD format.");
			op+= _T("\nx execute go		 Continue target execution.");
			op+= _T("\nuser              Attempt to display Forth USER area for suspended thread." );
			op+= _T("\ndasm <address>    Disassemble code starting at \"address\". ");
			op+= _T("\nua <address>      Disassemble code starting at \"address\". ");
			op+= _T("\ns step            Execute single instruction.");
			op+= _T(" \n");
			op+= _T(" \n");
			op+= _T(" \n");
	TextOut(op);
}

int CDebugConsole::CStringToInt(CString &c)
{
	int		acc = 0;
	int		radix = 10;
	int		sign = 1;
	TCHAR	ch;

	c.TrimLeft();
	if ( c.GetLength() == 0 ) return 0;
	if ( c[0] == '-' ) sign = -1;
	c.TrimLeft('-');

	c.TrimLeft('0');
	if ( c.GetLength() == 0 ) return 0;
	if ( c[0] == 'x' || c[0] == 'X' || c[0] == '$' ) radix = 16;
	c.TrimLeft(_T("xX$") );

	while (c.GetLength())
	{
		acc*=radix;
		ch = c[0];
		c = c.Right( c.GetLength() - 1 );
		switch( ch )
		{
		case '0':	acc+=0; break;
		case '1':	acc+=1; break;
		case '2':	acc+=2; break;
		case '3':	acc+=3; break;
		case '4':	acc+=4; break;
		case '5':	acc+=5; break;
		case '6':	acc+=6; break;
		case '7':	acc+=7; break;
		case '8':	acc+=8; break;
		case '9':	acc+=9; break;
		case 'a':	acc+=10; break;
		case 'b':	acc+=11; break;
		case 'c':	acc+=12; break;
		case 'd':	acc+=13; break;
		case 'e':	acc+=14; break;
		case 'f':	acc+=15; break;
		case 'A':	acc+=10; break;
		case 'B':	acc+=11; break;
		case 'C':	acc+=12; break;
		case 'D':	acc+=13; break;
		case 'E':	acc+=14; break;
		case 'F':	acc+=15; break;
		default:
			CString op = _T("Illegal expression.\n");
			TextOut( op );
			return 0;
		}
	}

	return acc;
}

void CDebugConsole::C_ContinueExec()
{
	CMainFrame *mf = (CMainFrame *)AfxGetMainWnd();
	mf->SendMessage( WM_COMMAND, ID_DEBUG_GO, 0 );
}

void CDebugConsole::C_User()
{
	CString op = _T("Not implemented yet.\n");
	TextOut( op );
}

void CDebugConsole::C_Dasm(CString &Command)
{
	CMainFrame *mf = (CMainFrame *)AfxGetMainWnd();

	int i = CStringToInt( Command );

	if ( mf->m_wndCodeView.IsWindowVisible() == 0 ) mf->ToggleWindowVisibility( &mf->m_wndCodeView );
	mf->m_wndCodeView.DisplayCode( (HANDLE)-1, (LPVOID)i );
}

void CDebugConsole::C_Step()
{
	CMainFrame *mf = (CMainFrame *)AfxGetMainWnd();
	mf->SendMessage( WM_COMMAND, ID_DEBUGGER_STEP, 0 );
}
