
// Project.cpp: implementation of the CProject class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "forge.h"
#include "Project.h"
#include "Targetproperties.h"
#include "mainfrm.h"
#include "projectreadstatus.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTarget::CTarget()
{
	m_szName = _T("<untitled target>");
	m_szPath = _T("");
	next = NULL;
	for (int i = 0 ; i<MAX_SOURCES_PER_TARGET; i++) Sources[i]=NULL;
	m_bOptimised = FALSE;
	m_bBuildType = 0;
	m_Platform = 0;
}

CTarget::~CTarget()
{
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CProject::CProject()
{
	SourceList = NULL;
	TargetList = NULL;
}

CProject::~CProject()
{
	if (SourceList!=NULL)
	{
		CSource	*n = SourceList->next;
		delete SourceList;
		SourceList = NULL;
		while (n!=NULL)
		{
			CSource *nn = n;
			n = n->next;
			delete nn;
		}
	}

	if (TargetList!=NULL)
	{
		CTarget	*n = TargetList->next;
		delete TargetList;
		TargetList = NULL;
		while (n!=NULL)
		{
			CTarget *nn = n;
			n = n->next;
			delete nn;
		}
	}
}

CSource * CProject::GetLastSourceEntry()
{
	CSource	*cs = SourceList;
	if (cs==NULL) return NULL;
	while (cs->next != NULL) cs=cs->next;
	return cs;
}

CTarget * CProject::GetLastTargetEntry()
{
	CTarget	*cs = TargetList;
	if (cs==NULL) return NULL;
	while (cs->next != NULL) cs=cs->next;
	return cs;
}

BOOL CProject::AddSource(LPCTSTR cs)
{
	CSource *os;
	CSource *ns = new CSource;
	ns->next = NULL;
	ns->m_szPath = cs;
	os = GetLastSourceEntry();
	if (os) os->next = ns;
	else SourceList = ns;
	return TRUE;
}

void CProject::SetBuildPath(LPCTSTR text)
{
	m_szBuildPath = text;
}

BOOL CProject::GenerateMakeFile( LPCTSTR targetname )
{
	CString tname = targetname;

	CTarget	*t = FindTargetByName( tname );
	ASSERT(t);

	CString basedir  = t->m_szPath;
	CString exename  = basedir+_T("\\");	exename+=t->m_szName;
	CString makefile = basedir+_T("\\makefile.vfx");
	CString	tooldir  = ( (CForgeApp*)AfxGetApp() )->m_BaseDirectory ;

	CStdioFile	*mf = new CStdioFile( makefile, CFile::modeCreate | CFile::modeWrite );

	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "( *T Makefile.VFX															)\n" ));
	mf->WriteString( _T( "( *P An automatically generated makefile from Forge						)\n" ));
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );

	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "( *S Generation options													)\n" ));
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );

	if ( t->m_bOptimised ) mf->WriteString( _T("optimised\n") ); else mf->WriteString( _T("unoptimised\n") );
	if ( t->m_bBuildType ) mf->WriteString( _T("DisableWin32Catch\t\\ Debugger gets exceptions for DEBUG build\n") );

	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "( *S Files to include														)\n" ));
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );
	
	for (int i = 0; i < MAX_SOURCES_PER_TARGET; i++)
	{
		if ( t->Sources[i] )
		{
			mf->WriteString( _T("S\" ") );
			mf->WriteString( t->Sources[i]->m_szPath );
			mf->WriteString( _T("\" INCLUDED \n") );
		}
	}

	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "( *S Binary SAVE operation												)\n" ));
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );
	mf->WriteString( _T( "save " ) );
	mf->WriteString( exename );
	mf->WriteString( _T( "\n" ) );

	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "( *S Build final EXE/DLL													)\n" ));
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );
	mf->WriteString( _T( "sh " ) );
	mf->WriteString( tooldir );
	mf->WriteString( _T("\\make-exe ") );
	mf->WriteString( exename );
	mf->WriteString( _T(".img ") );

	mf->WriteString( exename );
	mf->WriteString( _T(".exe ") );

	if (t->m_Platform == 0) mf->WriteString( _T("-G") );
	mf->WriteString( _T( "\n") );

	
	mf->WriteString( _T( "\n") );
	mf->WriteString( _T(  "\\ =======================================================================\n" ));
	mf->WriteString( _T( "\n") );

	mf->Close();

	return TRUE;
}

void CProject::Load(LPCTSTR szFileName)
{
	OPENFILENAME	ofn;

	CMainFrame *mf = (CMainFrame*)AfxGetMainWnd();

	if ( szFileName == NULL )
	{
		m_ProjectFile = _T("");

		ofn.lStructSize			=	sizeof( OPENFILENAME );
		ofn.hwndOwner			=	AfxGetMainWnd()->GetSafeHwnd();
		ofn.hInstance			=	AfxGetInstanceHandle();
		ofn.lpstrFilter			=	_T("Forge Workspace\0*.frg\0\0");
		ofn.lpstrCustomFilter	=	0;
		ofn.nMaxCustFilter		=	0;
		ofn.nFilterIndex		=	0;
		ofn.lpstrFile			=	m_ProjectFile.GetBuffer( 1024 );
		ofn.nMaxFile			=	1024;
		ofn.lpstrFileTitle		=	0;
		ofn.nMaxFileTitle		=	0;
		ofn.lpstrInitialDir		=	mf->m_csProjectDir;
		ofn.lpstrTitle			=	_T("Select Workspace...");
		ofn.Flags				=	OFN_PATHMUSTEXIST;
		ofn.nFileOffset			=	0;
		ofn.nFileExtension		=	0;
		ofn.lpstrDefExt			=	0;
		ofn.lCustData			=	0;
		ofn.lpfnHook			=	0;
		ofn.lpTemplateName		=	0;

		if (!::GetOpenFileName( &ofn ))
		{
			m_ProjectFile.ReleaseBuffer();
			return;
		}
	} else {
		m_ProjectFile = szFileName;
	}

	m_prsDialog = new CProjectReadStatus();
	m_prsDialog->Create( IDD_PROJECTREAD );

	CStdioFile	cf;
	CString	cs;
	CString command;

	CForgeApp *cforge = (CForgeApp *)AfxGetApp();
	cforge->m_MostRecentWorkspaceList->Add( m_ProjectFile );

	if ( cf.Open( m_ProjectFile, CFile::modeRead ) )
	{
		m_prsDialog->Status( _T("Workspace file open.") );
		while (cf.ReadString(cs))
		{
			command = cs.SpanExcluding( _T("\t \n\r") );
			command.MakeUpper();

			cs=cs.Right( cs.GetLength() - command.GetLength() );
			cs.TrimLeft();

			ProcessCommand( command, cs );
		}
		cf.Close();
	} else {
		m_prsDialog->Error( _T("Failed to open workspace file.") );
	}

	m_prsDialog->MakeModal();

	delete m_prsDialog;
}

void CProject::Save()
{
	OPENFILENAME	ofn;

	CMainFrame *mf = (CMainFrame*)AfxGetMainWnd();
	
	m_ProjectFile = _T("");

    ofn.lStructSize			=	sizeof( OPENFILENAME );
	ofn.hwndOwner			=	AfxGetMainWnd()->GetSafeHwnd();
    ofn.hInstance			=	AfxGetInstanceHandle();
	ofn.lpstrFilter			=	_T("Forge Workspace\0*.frg\0\0");
    ofn.lpstrCustomFilter	=	0;
	ofn.nMaxCustFilter		=	0;
    ofn.nFilterIndex		=	0;
	ofn.lpstrFile			=	m_ProjectFile.GetBuffer( 1024 );
    ofn.nMaxFile			=	1024;
	ofn.lpstrFileTitle		=	0;
    ofn.nMaxFileTitle		=	0;
	ofn.lpstrInitialDir		=	mf->m_csProjectDir;
    ofn.lpstrTitle			=	_T("Select Workspace Save Name");
	ofn.Flags				=	OFN_PATHMUSTEXIST;
    ofn.nFileOffset			=	0;
	ofn.nFileExtension		=	0;
    ofn.lpstrDefExt			=	_T("frg");
	ofn.lCustData			=	0;
    ofn.lpfnHook			=	0;
	ofn.lpTemplateName		=	0;

	if (!::GetSaveFileName( &ofn ))
	{
		m_ProjectFile.ReleaseBuffer();
		return;
	}

	CStdioFile	cf;

	CForgeApp *cforge = (CForgeApp *)AfxGetApp();
	cforge->m_MostRecentWorkspaceList->Add( m_ProjectFile );

	if ( cf.Open( m_ProjectFile, CFile::modeCreate | CFile::modeWrite ) )
	{

		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("# FORGE Project File\n") );
		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("#															 \n") );

		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("# BASE Information\n") );
		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("#															 \n") );

		// Date Stamp Project File
		cf.WriteString( _T("UpdatedOn\t\t") ); 
		CTime t = CTime::GetCurrentTime();
		CString ts = t.FormatGmt( _T("%#c") );
		ts+=_T("\n");
		cf.WriteString( ts );

		// Write Build Path to project File
		cf.WriteString( _T("BuildPath\t\t") ); 
		cf.WriteString( (LPCTSTR)m_szBuildPath );
		cf.WriteString( _T("\n") );

		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("# Source List\n") );
		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("#															 \n") );

		if (SourceList)
		{
			CSource		*cs = SourceList;
			while (cs!=NULL)
			{
				cf.WriteString( _T("Source___\t\t") );
				cf.WriteString( cs->m_szPath );
				cs=cs->next;
				cf.WriteString( _T("\n") );
				cf.WriteString( _T("#															 \n") );
			}
		}

		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("# Build Targets\n") );
		cf.WriteString( _T("# ###########################################################\n") );
		cf.WriteString( _T("#															 \n") );

		if (TargetList)
		{
			CTarget		*ct = TargetList;
			while (ct!=NULL)
			{
				cf.WriteString( _T("Target___\t\t") );
				cf.WriteString( ct->m_szName );	cf.WriteString( _T("\n") );

				cf.WriteString( _T("__PATH___\t\t") );
				cf.WriteString( ct->m_szPath );	cf.WriteString( _T("\n") );

				cf.WriteString( _T("__CODEGEN\t\t") );
				if ( ct->m_bBuildType ) cf.WriteString( _T("DEVELOPMENT\n") );
				else cf.WriteString( _T("PRODUCTION\n") );

				cf.WriteString( _T("__OPT____\t\t") );
				if ( ct->m_bOptimised ) cf.WriteString( _T("YES\n") );
				else cf.WriteString( _T("NO\n") );

				cf.WriteString( _T("__PLATFRM\t\t") );
				switch (ct->m_Platform)
				{
				case 0:
					cf.WriteString( _T("WIN32_GUI\n") );
					break;
				case 1:
					cf.WriteString( _T("WIN32_CUI\n") );
					break;
				case 2:
					cf.WriteString( _T("WIN32_DLL\n") );
					break;
				}

				for (int i=0;i<MAX_SOURCES_PER_TARGET;i++)
				{
					if ( ct->Sources[i]!=NULL )
					{
						CSource *c = ct->Sources[i];
						cf.WriteString( _T("__TSRC___\t\t") );
						cf.WriteString( c->m_szPath );	cf.WriteString( _T("\n") );
					}
				}
				cf.WriteString( _T("#															 \n") );
				ct=ct->next;
			}
		}


		cf.Close();
	}
}

BOOL CProject::ProcessCommand(CString& command, CString& arg)
{
	if ( command == _T("UPDATEDON") ) 
	{
		arg = "Last Updated on " + arg;
		m_prsDialog->Status( arg );
		return TRUE;
	}

	if ( command == _T("BUILDPATH") ) 
	{ 
		m_szBuildPath = arg; 
		arg = "Default Build Path is " + arg;
		m_prsDialog->Status( arg );
		return TRUE; 
	}

	if ( command == _T("SOURCE___") ) { AddSource(arg); return TRUE; }
	if ( command == _T("TARGET___") ) { AddTarget(arg,TRUE); return TRUE; }
	if ( command == _T("__PATH___") ) { SetLastTargetPath(arg); return TRUE; }
	if ( command == _T("__TSRC___") ) { AddSourceToLastTarget(arg); return TRUE; }
	if ( command == _T("__CODEGEN") ) { SetTargetCodeGen(arg); return TRUE; }
	if ( command == _T("__PLATFRM") ) { SetTargetPlatform(arg); return TRUE; }
	if ( command == _T("__OPT____") ) { SetTargetOptState(arg); return TRUE; }
	if ( command == _T("#") ) return TRUE;

	CString text = "Ignoring unrecognized command '" + arg + "' in workspace.";
	m_prsDialog->Warning( text );

	return FALSE;
}


BOOL CProject::DoesSourceExist(LPCTSTR path)
{
	CString	name = path;
	if ( FindSourceByName( name ) ) return TRUE;
	return FALSE;
}

BOOL CProject::DoesTargetExist(LPCTSTR path)
{
	CString	name = path;
	if ( FindTargetByName( name ) ) return TRUE;
	return FALSE;
}

void CProject::RemoveSourceByName( CString& name )
{
	if (SourceList)
	{
		CSource		*cs   = SourceList;
		CSource		*prev = NULL;
		while (cs!=NULL)
		{
			if ( cs->m_szPath == name ) 
			{
				// I want to remove this entry
				if ( prev == NULL )
				{
					SourceList = cs->next;
				} else {
					prev->next =cs->next;
				}
				delete cs;
				return;
			}
			prev = cs;
			cs=cs->next;
		}
	}
}

BOOL CProject::IsSourceInUse( CString& name )
{
	if (TargetList)
	{
		CTarget		*cs = TargetList;
		while (cs!=NULL)
		{
			for ( int i = 0; i < MAX_SOURCES_PER_TARGET; i++ )
			{
				CSource *cursource = cs->Sources[i];
				if (cursource)
					if ( cursource->m_szPath == name ) return TRUE;
			}
			cs = cs->next;
		}
	}

	return FALSE;
}

BOOL CProject::AddTarget(LPCTSTR name, BOOL bFromFile )
{
	CTarget *os;
	CTarget *ns = new CTarget;
	ns->next = NULL;
	ns->m_szPath = _T("");
	ns->m_szName = name;
	os = GetLastTargetEntry();
	if (os) os->next = ns;
	else TargetList = ns;

	if (!bFromFile) 
	{
		EditTargetProperties( NULL );
	}

	return TRUE;
}

void CProject::SetLastTargetPath(LPCTSTR path)
{
	GetLastTargetEntry()->m_szPath = path;
}

void CProject::AddSourceToTarget(CTarget *targ, LPCTSTR source)
{
	if ( targ == NULL ) targ = GetLastTargetEntry();

	int i = targ->GetNextFreeSourceSlot();
	if ( i == -1 ) ::MessageBox( HWND_DESKTOP, "Maximum number of sources per target exceeded.", _T("Forge Error"), MB_OK | MB_ICONEXCLAMATION );
	else targ->Sources[i] = FindSourceByName( source );
}

void CProject::AddSourceToLastTarget(LPCSTR sourcename)
{
	AddSourceToTarget( NULL, sourcename );
}

void CProject::EditTargetProperties(LPCTSTR name)
{
	CString	tName = name;

	CTargetProperties *tp = new CTargetProperties();

	if ( name == NULL )
	{
		tp->m_Target = GetLastTargetEntry();
	} else {
		tp->m_Target = FindTargetByName( tName );
		if ( tp->m_Target == NULL ) return;
	}
	
	tp->DoModal();
	delete tp;
}

CTarget * CProject::FindTargetByName(CString &name)
{
	if (TargetList)
	{
		CTarget		*cs = TargetList;
		while (cs!=NULL)
		{
			if ( cs->m_szName == name ) return cs;
			cs=cs->next;
		}
	}
	return NULL;
}

CSource * CProject::FindSourceByName( LPCTSTR name)
{
	if (SourceList)
	{
		CSource		*cs = SourceList;
		while (cs!=NULL)
		{
			if ( cs->m_szPath == name ) return cs;
			cs=cs->next;
		}
	}
	return NULL;
}

int CTarget::GetNextFreeSourceSlot()
{
	for ( int i = 0; i<MAX_SOURCES_PER_TARGET; i++ )
		if ( Sources[i]==NULL ) return i;
	return -1;
}

void CProject::SetTargetCodeGen(LPCTSTR type)
{
	CString test = type;
	
	if ( test.CompareNoCase( _T("PRODUCTION") ) == 0 )  
	{
		GetLastTargetEntry()->m_bBuildType = FALSE;
		return;
	}

	if ( test.CompareNoCase( _T("DEVELOPMENT") ) == 0 ) 
	{
		GetLastTargetEntry()->m_bBuildType = TRUE;
		return;
	}

	test = _T("Unknown Code Generation Type : ") + test;
	m_prsDialog->Warning( test );
}


void CProject::SetTargetPlatform(LPCTSTR arg)
{
	CString test = arg;
	
	if ( test.CompareNoCase( _T("WIN32_GUI") ) == 0 )  
	{
		GetLastTargetEntry()->m_Platform = 0;
		return;
	}

	if ( test.CompareNoCase( _T("WIN32_CUI") ) == 0 )  
	{
		GetLastTargetEntry()->m_Platform = 1;
		return;
	}

	if ( test.CompareNoCase( _T("WIN32_DLL") ) == 0 )  
	{
		GetLastTargetEntry()->m_Platform = 2;
		return;
	}

	test = _T("Unknown target platform: ") + test;
	m_prsDialog->Warning( test );
}

void CProject::SetTargetOptState(LPCTSTR arg)
{
	CString test = arg;
	
	if ( test.CompareNoCase( _T("YES") ) == 0 )  
	{
		GetLastTargetEntry()->m_bOptimised = TRUE;
		return;
	}

	if ( test.CompareNoCase( _T("NO") ) == 0 ) 
	{
		GetLastTargetEntry()->m_bOptimised = FALSE;
		return;
	}

	test = _T("Unknown Optimiser Mode: ") + test;
	m_prsDialog->Warning( test );
}

void CTarget::RemoveSourceFromTarget(int index)
{
	for ( int i = index; i < (MAX_SOURCES_PER_TARGET-1); i++ )
	{
		Sources[i] = Sources[i+1];
	}
	Sources[ MAX_SOURCES_PER_TARGET-1 ] = NULL;
}

void CTarget::PromoteTargetSource(int index)
{
	if ( index > 0 )
	{
		CSource *t = Sources[ index ];
		Sources[index] = Sources[ index-1 ];
		Sources[index-1] = t;
	}
}

void CTarget::DemoteTargetSource(int index)
{
	CSource *t = Sources[ index+1 ];
	Sources[index+1] = Sources[ index ];
	Sources[index] = t;
}
