// TSEPOS 8 1 7 //
// ProjectWindow.cpp: implementation of the CProjectWindow class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "forge.h"
#include "project.h"
#include "mainfrm.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

IMPLEMENT_DYNAMIC(CProjectWindow, CSizingControlBar)

BEGIN_MESSAGE_MAP(CProjectWindow, CSizingControlBar)
	//{{AFX_MSG_MAP(CProjectWindow)
	ON_WM_SIZE()
	ON_WM_CONTEXTMENU()
	ON_COMMAND(ID_PROJECT_SOURCE_OPENSELECTED, OnProjectSourceOpenselected)
	ON_UPDATE_COMMAND_UI(ID_PROJECT_SOURCE_OPENSELECTED, OnUpdateProjectSourceOpenselected)
	ON_UPDATE_COMMAND_UI(ID_PROJECT_SOURCE_REMOVESELECTED, OnUpdateProjectSourceRemoveselected)
	ON_COMMAND(ID_PROJECT_SOURCE_REMOVESELECTED, OnProjectSourceRemoveselected)
	ON_COMMAND(ID_TARGET_PROPERTIES, OnTargetProperties)
	ON_COMMAND(ID_TARGET_DELETE, OnTargetDelete)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CProjectWindow::CProjectWindow()
{
}

CProjectWindow::~CProjectWindow()
{

}

BOOL CProjectWindow::Create(LPCTSTR lpszWindowName, CWnd * pParentWnd, CSize sizeDefault, BOOL bHasGripper)
{
	CSizingControlBar::Create(lpszWindowName, pParentWnd, sizeDefault, bHasGripper, IDR_PROJECTWINDOW );
	m_Tree.Create( WS_CHILD | TVS_HASLINES | TVS_LINESATROOT | TVS_HASBUTTONS ,
					CRect(0,0,0,0), this, IDR_PROJECTTREE );
	m_Tree.ModifyStyleEx( 0, WS_EX_CLIENTEDGE, 0 );
	ht_Sources = m_Tree.InsertItem( _T("Sources") );
	ht_Targets = m_Tree.InsertItem( _T("Targets") );
	return TRUE;
}

void CProjectWindow::ResetProjectWindow(BOOL bProjectOpen)
{
	if (!bProjectOpen) { m_Tree.ShowWindow( SW_HIDE ); return; }
	UpdateSourceBranch();
	UpdateTargetBranch();
	m_Tree.ShowWindow( SW_NORMAL );
	m_Tree.InvalidateRect( NULL );

	// Now rebuild the active targets control
	CMainFrame *mf = (CMainFrame *)AfxGetMainWnd();
	CDialogBar *cd = &(mf->m_wndActiveTarget);
	CComboBox  *cb = (CComboBox *)cd->GetDlgItem( IDC_TARGETS );
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;

	CString	oselect = _T("");
	int i = cb->GetCurSel(); 
	if ( i!=CB_ERR ) cb->GetLBText( i, oselect );

	cb->ResetContent();

	CTarget	*cs = p->TargetList;
	while (cs!=NULL)
	{
		cb->AddString( cs->m_szName );
		cs=cs->next;
	}

	i = cb->FindStringExact( 0, oselect );
	if ( i==CB_ERR ) i = 0;
	cb->SetCurSel( i );
}

void CProjectWindow::OnSize(UINT nType, int cx, int cy)
{
	if (IsWindow( m_Tree ) )	m_Tree.MoveWindow( 0, 0, cx, cy );
}


BOOL TreeChildExists( CTreeCtrl *tree, HTREEITEM branch, LPCTSTR text )
{
	HTREEITEM	child = tree->GetChildItem(branch);
	while (child)
	{
		CString stext = tree->GetItemText( child );
		if ( stext == text ) return TRUE;
		child = tree->GetNextSiblingItem( child );
	}
	return FALSE;
}

void CProjectWindow::UpdateSourceBranch()
{
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;

	// Step #1: Enumerate project sources and add to tree any item which does not yet
	//			exist.

	if (p->SourceList)
	{
		CSource		*cs = p->SourceList;
		while (cs!=NULL)
		{
			if ( !::TreeChildExists( &m_Tree, ht_Sources, cs->m_szPath ) )
				m_Tree.InsertItem( cs->m_szPath, ht_Sources );
			cs=cs->next;
		}
	}

	// Step #2: Walk tree items and remove any which cannot be found in project

	HTREEITEM	child = m_Tree.GetChildItem(ht_Sources);
	while (child)
	{
		CString stext = m_Tree.GetItemText( child );
		HTREEITEM next = m_Tree.GetNextSiblingItem( child );
		if (!p->DoesSourceExist( stext )) m_Tree.DeleteItem(child);
		child = next;
	}

}

void CProjectWindow::UpdateTargetBranch()
{
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;

	// Step #1: Enumerate project sources and add to tree any item which does not yet
	//			exist.

	if (p->SourceList)
	{
		CTarget	*cs = p->TargetList;
		while (cs!=NULL)
		{
			if ( !::TreeChildExists( &m_Tree, ht_Targets, cs->m_szName ) )
				m_Tree.InsertItem( cs->m_szName, ht_Targets );
			cs=cs->next;
		}
	}

	// Step #2: Walk tree items and remove any which cannot be found in project

	HTREEITEM	child = m_Tree.GetChildItem(ht_Targets);
	while (child)
	{
		CString stext = m_Tree.GetItemText( child );
		HTREEITEM next = m_Tree.GetNextSiblingItem( child );
		if (!p->DoesTargetExist( stext )) m_Tree.DeleteItem(child);
		child = next;
	}

}


void CProjectWindow::OnContextMenu(CWnd* pWnd, CPoint point) 
{
	int MenuX = point.x;
	int MenuY = point.y;

	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;

	if (p==NULL) return;					// No project.. no menu simple eh?
	if (pWnd != (CWnd*)&m_Tree)	return;		// Not from the tree view so ....

	UINT flag;
	m_Tree.ScreenToClient(&point);
	HTREEITEM	hItem = m_Tree.HitTest( point, &flag );

	if ( ( flag & TVHT_ONITEM ) == 0 ) return;

	m_Tree.SelectItem( hItem );
	CString name = m_Tree.GetItemText( hItem );


	// If item is HT_SOURCES then menu is for SOURCES root
	if (hItem == ht_Sources) 
	{
		CMenu	*cm = new CMenu();
		cm->LoadMenu( IDR_PROJECT_CONTEXTMENUS );
		cm->GetSubMenu( 0 )->TrackPopupMenu( TPM_LEFTALIGN, MenuX, MenuY, this );
		delete cm;
		return;
	}

	// If parent item is HT_SOURCES then menu is for an individual source file
	if ( m_Tree.GetParentItem( hItem ) == ht_Sources)
	{
		CMenu	*cm = new CMenu();
		cm->LoadMenu( IDR_PROJECT_CONTEXTMENUS );
		cm->GetSubMenu( 1 )->TrackPopupMenu( TPM_LEFTALIGN, MenuX, MenuY, this );
		delete cm;
		return;
	}

	// If item is HT_TARGETS then menu is for a target-target root
	if ( hItem == ht_Targets)
	{
		CMenu	*cm = new CMenu();
		cm->LoadMenu( IDR_PROJECT_CONTEXTMENUS );
		cm->GetSubMenu( 2 )->TrackPopupMenu( TPM_LEFTALIGN, MenuX, MenuY, this );
		delete cm;
		return;
	}

	// If parent item is HT_TARGETS then menu is for a target root
	if ( m_Tree.GetParentItem( hItem ) == ht_Targets)
	{
		CMenu	*cm = new CMenu();
		cm->LoadMenu( IDR_PROJECT_CONTEXTMENUS );
		cm->GetSubMenu( 3 )->TrackPopupMenu( TPM_LEFTALIGN, MenuX, MenuY, this );
		delete cm;
		return;
	}

}

void CProjectWindow::OnProjectSourceOpenselected() 
{
	POSITION p = AfxGetApp()->GetFirstDocTemplatePosition();
	CDocTemplate * doctemplate = AfxGetApp()->GetNextDocTemplate( p );
	CString name = m_Tree.GetItemText( m_Tree.GetSelectedItem() );
	doctemplate->OpenDocumentFile( name, TRUE );
}

void CProjectWindow::OnUpdateProjectSourceOpenselected(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable();
}


BOOL CProjectWindow::OnNotify(WPARAM wParam, LPARAM lParam, LRESULT* pResult) 
{

	NMHDR	*h = (NMHDR *)lParam;

	if ( ( h->hwndFrom == (HWND)m_Tree ) && ( h->code == NM_DBLCLK ) )
	{
		if ( m_Tree.GetParentItem( m_Tree.GetSelectedItem() ) == ht_Sources) OnProjectSourceOpenselected();
		if ( m_Tree.GetParentItem( m_Tree.GetSelectedItem() ) == ht_Targets ) OnTargetProperties();
	}

	return CSizingControlBar::OnNotify(wParam, lParam, pResult);
}

void CProjectWindow::OnUpdateProjectSourceRemoveselected(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable();
}

void CProjectWindow::OnProjectSourceRemoveselected() 
{
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;
	CString name = m_Tree.GetItemText( m_Tree.GetSelectedItem() );

	if ( p->IsSourceInUse( name ) )
	{
		MessageBox( _T("The selected cannot be removed whilst it is being used in at least one target. Remove source from all target builds before delete."),NULL,MB_OK|MB_ICONINFORMATION );
		return;
	}
	p->RemoveSourceByName( name );
	ResetProjectWindow( TRUE );
}

void CProjectWindow::OnTargetProperties() 
{
	CString name; GetSelectedTargetName( name ); 
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;
	p->EditTargetProperties( name );	
	ResetProjectWindow(TRUE);
}

void CProjectWindow::GetSelectedTargetName(CString &c)
{
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;
	c = m_Tree.GetItemText( m_Tree.GetSelectedItem() );
}

void CProjectWindow::OnTargetDelete() 
{
	::MessageBox( HWND_DESKTOP, "CProjectWindow::OnTargetDelete()", "TODO", MB_OK ) ;	
}
