// TargetProperties.cpp : implementation file
//

#include "stdafx.h"
#include "forge.h"
#include "TargetProperties.h"
#include "project.h"
#include "selectsource.h"
#include "mainfrm.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CTargetProperties dialog


CTargetProperties::CTargetProperties(CWnd* pParent /*=NULL*/)
	: CDialog(CTargetProperties::IDD, pParent)
{
	//{{AFX_DATA_INIT(CTargetProperties)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void CTargetProperties::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CTargetProperties)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CTargetProperties, CDialog)
	//{{AFX_MSG_MAP(CTargetProperties)
	ON_BN_CLICKED(IDC_ADD, OnAdd)
	ON_BN_CLICKED(IDC_REMOVE, OnRemove)
	ON_BN_CLICKED(IDC_PROMOTE, OnPromote)
	ON_BN_CLICKED(IDC_DEMOTE, OnDemote)
	ON_LBN_SELCANCEL(IDC_SOURCELIST, OnSelcancelSourcelist)
	ON_LBN_SELCHANGE(IDC_SOURCELIST, OnSelchangeSourcelist)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CTargetProperties message handlers

BOOL CTargetProperties::OnInitDialog() 
{
	CDialog::OnInitDialog();

	SetDlgItemText( IDC_TARGET_NAME, m_Target->m_szName );
	SetDlgItemText( IDC_TARGET_PATH, m_Target->m_szPath );
	if ( m_Target->m_bBuildType )
		CheckRadioButton( IDC_BUILD_PRODUCTION, IDC_BUILD_DEVELOPMENT, IDC_BUILD_DEVELOPMENT );
	else
		CheckRadioButton( IDC_BUILD_PRODUCTION, IDC_BUILD_DEVELOPMENT, IDC_BUILD_PRODUCTION );

	CheckRadioButton( IDC_GUI, IDC_DLL, IDC_GUI+m_Target->m_Platform );

	if ( m_Target->m_bOptimised ) CheckDlgButton( IDC_OPTIMISED, 1 );

	RefreshSourceList();

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CTargetProperties::OnOK() 
{

	// Read target name into temporary string and check for .EXE on the end, if found the warn and refuse OK


	CString	temp;
	GetDlgItemText( IDC_TARGET_NAME, temp );
	temp.MakeUpper();
	if ( temp.Right(4) == _T(".EXE") ) 
	{
		::MessageBox( HWND_DESKTOP, "You should not specify the EXE file extension to the target name.", "Forge Error (Target Properties)", MB_OK | MB_ICONERROR );
		return;
	}

	
	// Should check that m_szName is unique (other than self) and that PATH exists!

	GetDlgItemText( IDC_TARGET_NAME, m_Target->m_szName );
	GetDlgItemText( IDC_TARGET_PATH, m_Target->m_szPath );
	m_Target->m_bBuildType = (BOOL)IsDlgButtonChecked( IDC_BUILD_DEVELOPMENT );

	if (IsDlgButtonChecked( IDC_GUI ) ) m_Target->m_Platform = 0;
	if (IsDlgButtonChecked( IDC_CONSOLE ) ) m_Target->m_Platform = 1;
	if (IsDlgButtonChecked( IDC_DLL ) ) m_Target->m_Platform = 2;
	m_Target->m_bOptimised = (BOOL)IsDlgButtonChecked( IDC_OPTIMISED );

	CDialog::OnOK();
}

void CTargetProperties::AddItem(LPCTSTR text)
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	lb->AddString( text );
}

void CTargetProperties::RefreshSourceList()
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	lb->ResetContent();
	for (int i = 0; i<MAX_SOURCES_PER_TARGET;i++ )
		if ( m_Target->Sources[i] ) AddItem( m_Target->Sources[i]->m_szPath );
}

void CTargetProperties::OnAdd() 
{
	CProject *p = ((CMainFrame *)AfxGetMainWnd())->m_Project;
 	CSelectSource *s = new CSelectSource();
 	s->SourceList[0] = p->SourceList;
 	s->b_SourceListIsArray = FALSE;
 	s->DoModal();

	if ( s->SelectedName.GetLength() > 0 )
 	{
		p->AddSourceToTarget( m_Target , s->SelectedName );
		RefreshSourceList();
 	}
	delete s;

	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	lb->SetCurSel( lb->GetCount()-1 );
}

void CTargetProperties::OnRemove() 
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	int i = lb->GetCurSel();
	if ( i != LB_ERR ) 
	{
		m_Target->RemoveSourceFromTarget( i );
		RefreshSourceList();
		if ( i !=0 ) lb->SetSel( i+1 );
		else lb->SetCurSel(i);
	}
}

void CTargetProperties::OnPromote() 
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	int i = lb->GetCurSel();
	if ( i != LB_ERR ) 
	{
		m_Target->PromoteTargetSource( i );
		RefreshSourceList();
		lb->SetCurSel( i-1 );
	}
}

void CTargetProperties::OnDemote() 
{
	CListBox *lb = (CListBox *)GetDlgItem(IDC_SOURCELIST);
	int i = lb->GetCurSel();
	if ( i != LB_ERR ) 
	{
		if ( i < lb->GetCount()-1 )
		{
			m_Target->DemoteTargetSource( i );
			RefreshSourceList();
			lb->SetCurSel( i+1 );
		}
	}
}

void CTargetProperties::OnSelcancelSourcelist() 
{
}

void CTargetProperties::OnSelchangeSourcelist() 
{
	GetDlgItem( IDC_PROP )->EnableWindow();
	GetDlgItem( IDC_REMOVE )->EnableWindow();
	GetDlgItem( IDC_PROMOTE )->EnableWindow();
	GetDlgItem( IDC_DEMOTE )->EnableWindow();
}
