// ToolCapture.cpp : implementation file
//

#include "stdafx.h"
#include "ToolCapture.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CToolCapture

CToolCapture::CToolCapture()
{
	m_uTimer = 0;
	hChildProcess = NULL;
	bTerminateBanner = TRUE;
	bAutoClearOnStart = TRUE;
}

CToolCapture::~CToolCapture()
{
}


BEGIN_MESSAGE_MAP(CToolCapture, CRichEditCtrl)
	//{{AFX_MSG_MAP(CToolCapture)
	ON_WM_CREATE()
	ON_WM_DESTROY()
	ON_WM_TIMER()
	ON_WM_CHAR()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CToolCapture message handlers

int CToolCapture::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CRichEditCtrl::OnCreate(lpCreateStruct) == -1)
		return -1;

	m_Font.CreatePointFont( 100, "Courier", NULL );
	SetFont(&m_Font);
	
//	EnableWindow(FALSE);

	return 0;
}

void CToolCapture::OnDestroy() 
{
	CRichEditCtrl::OnDestroy();
	DestroyPipes();
}

int CToolCapture::LaunchProcess(LPCTSTR szProcess, LPCTSTR szStartDirectory)
{
	if (hChildProcess) return TCERR_PROCESSRUNNING;

	DestroyPipes(); if (!CreatePipes()) return TCERR_NOPIPES;

//	EnableWindow(TRUE);

	m_uTimer = SetTimer( 1, 100, NULL );
	if (m_uTimer == 0) return TCERR_NOTIMER;


	STARTUPINFO			sui;
	PROCESS_INFORMATION	pi;

	::ZeroMemory( &sui, sizeof(STARTUPINFO) );
	sui.cb = sizeof(STARTUPINFO);

	sui.dwFlags			=	STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
	sui.hStdInput		=	hRemStdin;
	sui.hStdOutput		=	hRemStdout;
	sui.hStdError 		=	hRemStderr;
	sui.wShowWindow		=	SW_HIDE;

	::SetCurrentDirectory( szStartDirectory );

	if (!CreateProcess( NULL, (LPSTR)szProcess, NULL, NULL, TRUE, NULL, NULL, szStartDirectory, &sui, &pi ) )
	{
		ChildFinished();
		return TCERR_FAILEDLAUNCH;
	}

	if (bAutoClearOnStart) ClearScreen();

	hChildProcess = pi.hProcess;
	hChildThread  = pi.hThread;
    dwProcessId   = pi.dwProcessId;
	dwThreadId	  = pi.dwThreadId;

	return 0;
}

void CToolCapture::ChildFinished()
{
	if (m_uTimer) KillTimer(m_uTimer);
	m_uTimer = 0;

	dwProcessId = dwThreadId = 0;
	hChildProcess = hChildThread = NULL;
	DestroyPipes();
	WriteToControl("\r\nCapture terminated\r\n");
//	EnableWindow(FALSE);
}

void CToolCapture::DestroyPipes()
{
	if ( hLocStdin  ) { CloseHandle( hLocStdin  ); hLocStdin  = NULL; }
	if ( hLocStdout ) { CloseHandle( hLocStdout ); hLocStdout = NULL; }
	if ( hLocStderr ) { CloseHandle( hLocStderr ); hLocStderr = NULL; }
	if ( hRemStdin  ) { CloseHandle( hRemStdin  ); hRemStdin  = NULL; }
	if ( hRemStdout ) { CloseHandle( hRemStdout ); hRemStdout = NULL; }
	if ( hRemStderr ) { CloseHandle( hRemStderr ); hRemStderr = NULL; }
}


BOOL CToolCapture::CreatePipes()
{

	SECURITY_ATTRIBUTES	sattr;
	sattr.nLength			=	sizeof(SECURITY_ATTRIBUTES);
	sattr.lpSecurityDescriptor = NULL;
	sattr.bInheritHandle	=	TRUE;

	if ( !::CreatePipe( &hRemStdin,  &hLocStdin,  &sattr, 1024 ) ) return FALSE;
	if ( !::CreatePipe( &hLocStdout, &hRemStdout, &sattr, 1024 ) ) return FALSE;
	if ( !::CreatePipe( &hLocStderr, &hRemStderr, &sattr, 1024 ) ) return FALSE;
	return TRUE;
}

void CToolCapture::OnTimer(UINT nIDEvent) 
{
	DWORD	ec;
	if (hChildProcess!=NULL)
	{
		if (::PeekNamedPipe( hLocStdout, NULL, NULL, NULL, &ec, NULL) ) PipeToControl(hLocStdout,ec);
		if (::PeekNamedPipe( hLocStderr, NULL, NULL, NULL, &ec, NULL) ) PipeToControl(hLocStderr,ec);

		::GetExitCodeProcess( hChildProcess, &ec );
		if ( ec != STILL_ACTIVE ) ChildFinished();
	}
}


void CToolCapture::PipeToControl( HANDLE pipe, DWORD numbytes)
{
	void	*p;
	DWORD	bytesdone;

	if (numbytes)
	{
		p = ::GlobalAlloc( GPTR, numbytes+1 );
		if (p)
		{
			::ReadFile( pipe, p, numbytes, &bytesdone, NULL );
			WriteToControl( (LPSTR)p );	
			::GlobalFree(p);
		}
	}
}

void CToolCapture::WriteToControl(LPSTR text)
{
	SetSel( -1, -1 );
	ReplaceSel(text);
	SendMessage(EM_SCROLLCARET);
}

void CToolCapture::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// TODO: Add your message handler code here and/or call default
	char	c;
	DWORD	bytesdone;

	if (hLocStdin!=NULL)
	{
		c = (char)nChar;
		::WriteFile( hLocStdin, &c, sizeof(char), &bytesdone, NULL );
	}
}

int CToolCapture::TerminateCapture()
{
	if (hChildProcess==NULL) return 0;
	if ( ::TerminateProcess( hChildProcess, -1 ) ) return 0;
	return TCERR_CANNOTERMINATE;
}

void CToolCapture::ClearScreen()
{
	SetWindowText(" ");
}


BOOL CToolCapture::Create( const RECT& cr, CWnd * parent )
{
	return CRichEditCtrl::Create( WS_CHILD | WS_VISIBLE | WS_BORDER | ES_MULTILINE | ES_WANTRETURN |
					ES_AUTOHSCROLL | WS_HSCROLL | WS_VSCROLL | ES_AUTOVSCROLL,
					cr, parent, 1 );

}




